from dataclasses import dataclass
from datetime import timedelta
from typing import Dict, Optional
import abc
import hashlib
import logging
import os
import pathlib
import pickle
import sqlite3


@dataclass
class CacheInfo:
    """Information about a cache. Note that the information can be approximated
    as getting the exact numbers may be costly."""

    size: int = 0
    """Approximate number of bytes stored in the cache."""

    entry_count: int = 0
    """Approximate number of objects stored in the cache."""

    name: str = ""
    """A human-friendly name for this cache."""


class Cache(abc.ABC):
    
    """Interface for key-value caches.
    """
    @abc.abstractmethod
    def set_key_prefix(self, prefix: bytes):
        """Set a prefix for all keys. This can be for example used to
        avoid collisions when the configuration changes (which can
        impact all values)."""
        ...

    @abc.abstractmethod
    def put(self, key: bytes, value: object) -> bool:
        """Put a value into the cache using the provided key.

        :param key: The key under which ``value`` will be stored.
        :param value: A pickable Python object to be stored.
        :return: ``True`` if the value was added to the cache, ``False`` if
                 it was already cached.
        """

    @abc.abstractmethod
    def get(self, key: bytes) -> Optional[object]:
        """Get a stored object.

        :param key: The object key.
        :return: An object if one exists. Otherwise, return ``None``.
        """

    def persist(self) -> None:
        """Persists this cache to disk/persistent storage.

        This function should be called after the cache has been populated. On
        the next run, the constructor will then pick up the index and return
        cached data.
        """
        return None

    @abc.abstractmethod
    def clear(self) -> None:
        """Clear the contents of the cache.

        .. versionadded:: 2.5
        """

    @abc.abstractmethod
    def inspect(self) -> CacheInfo:
        """Get an overview of the cached data.

        .. versionadded:: 2.5"""


class _CacheBase(Cache):
    """A base class which handles the 
    :py:meth:`~liara.cache.Cache.set_key_prefix` transparently."""

    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self):
        self.__prefix = bytes()

    def set_key_prefix(self, prefix: bytes):
        self.__log.debug('Setting cache key prefix to: "%s"',
                         prefix.hex())
        self.__prefix = prefix

    @abc.abstractmethod
    def _put(self, key: bytes, value: object) -> bool:
        """To be implemented by derived classes, this method will be called
        with the prefix already applied."""
        pass

    @abc.abstractmethod
    def _get(self, key: bytes) -> Optional[object]:
        """To be implemented by derived classes, this method will be called
        with the prefix already applied."""
        pass

    def put(self, key: bytes, value: object) -> bool:
        return self._put(self.__prefix + key, value)
    
    def get(self, key: bytes) -> Optional[object]:
        return self._get(self.__prefix + key)


class FilesystemCache(_CacheBase):
    """A :py:class:`Cache` implementation which uses the filesystem to cache
    data.

    This cache tries to load a previously generated index. Use
    :py:meth:`persist` to write the cache index to disk.
    """
    __index: Dict[bytes, pathlib.Path]

    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, path: pathlib.Path):
        super().__init__()
        self.__path = path
        self.__index = {}
        os.makedirs(self.__path, exist_ok=True)
        self.__index_file = self.__path / 'cache.index'
        if self.__index_file.exists():
            try:
                self.__index = pickle.load(self.__index_file.open('rb'))
            except Exception:
                # Not being able to load the cache is not an error
                pass

    def clear(self) -> None:
        import shutil
        self.__log.debug('Clearing cache')
        self.__index = {}
        shutil.rmtree(self.__path, ignore_errors=True)
        os.makedirs(self.__path, exist_ok=True)

    def persist(self):
        pickle.dump(self.__index, self.__index_file.open('wb'))

    def _put(self, key: bytes, value: object) -> bool:
        if key in self.__index:
            return False

        cache_object_path = self.__path / hashlib.sha256(key).hexdigest()
        pickle.dump(value, cache_object_path.open('wb'))

        self.__index[key] = cache_object_path
        return True

    def _get(self, key: bytes) -> Optional[object]:
        if key not in self.__index:
            return None

        cache_object_path = self.__index[key]
        return pickle.load(cache_object_path.open('rb'))

    def inspect(self):
        count = len(self.__index)

        size = 0
        for path in self.__index.values():
            size += pathlib.Path(path).stat().st_size

        return CacheInfo(size, count, 'Filesystem')


class Sqlite3Cache(_CacheBase):
    """A :py:class:`Cache` implementation which uses SQLite to store the data.
    This is mostly useful if creating many files is slow, for instance due to
    anti-virus software.

    This cache tries to load a previously generated index. Use
    :py:meth:`persist` to write the cache index to disk.
    """

    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, path: pathlib.Path):
        super().__init__()
        self.__path = path
        os.makedirs(self.__path, exist_ok=True)
        self.__db_file = self.__path / 'cache.db'
        # We're only ever used by one thread, but we may get passed around
        # from one thread to another when running `serve`.
        self.__connection = sqlite3.connect(self.__db_file, check_same_thread=False)
        self.__cursor = self.__connection.cursor()
        self.__cursor.execute("""CREATE TABLE IF NOT EXISTS cache
            (key BLOB PRIMARY KEY NOT NULL,
                data BLOB NOT NULL,
                object_type VARCHAR NOT NULL);""")

    def clear(self) -> None:
        self.__log.debug('Clearing cache')
        self.__cursor.execute("DELETE FROM cache;")
        self.__connection.commit()
        self.__cursor.execute("VACUUM;")

    def persist(self):
        self.__connection.commit()

    def _put(self, key: bytes, value: object) -> bool:
        # The semantics are such that inserting the same key twice should not
        # cause a failure, so we ignore failures here
        q = 'INSERT OR IGNORE INTO cache VALUES(?, ?, ?);'

        # We check for byte(array), image nodes for instance store binary data
        # directly, and there's no need to send it through pickle. It doesn't
        # seem to make much of measurable difference though
        if isinstance(value, bytes) or isinstance(value, bytearray):
            object_type = 'BINARY'
        else:
            object_type = 'OBJECT'
            value = pickle.dumps(value)

        r = self.__cursor.execute(q, (key, value, object_type,))

        return r.lastrowid != 0

    def _get(self, key: bytes) -> Optional[object]:
        q = 'SELECT data, object_type FROM cache WHERE key=?'
        self.__cursor.execute(q, (key,))
        r = self.__cursor.fetchone()

        if r:
            if r[1] == 'OBJECT':
                return pickle.loads(r[0])
            else:
                return r[0]

        return None

    def inspect(self):
        size, count = self.__cursor.execute(
            'SELECT SUM(LENGTH(data)), COUNT(*) FROM cache;'
        ).fetchone()

        # If there is no data stored, the query above will return None, None
        if size is None:
            size = 0

        if count is None:
            count = 0

        return CacheInfo(size, count, 'Sqlite3')


class MemoryCache(_CacheBase):
    """An in-memory :py:class:`Cache` implementation.

    This cache stores all objects in-memory.
    """
    __index: Dict[bytes, object]

    def __init__(self):
        super().__init__()
        self.__index = {}

    def _put(self, key: bytes, value: object) -> bool:
        if key in self.__index:
            return False

        self.__index[key] = value
        return True

    def _get(self, key: bytes) -> Optional[object]:
        return self.__index.get(key, None)

    def clear(self):
        self.__index = {}

    def inspect(self):
        import sys
        size = sys.getsizeof(self.__index)

        size += sum([sys.getsizeof(k) + sys.getsizeof(v)
                     for k, v in self.__index.items()])

        count = len(self.__index)

        return CacheInfo(size, count, 'In-Memory')


class RedisCache(_CacheBase):
    """A cache using Redis as the storage backend."""

    def __init__(self, host: str, port: int, db: int,
                 expiration_time=timedelta(hours=1)):
        import redis
        super().__init__()
        self.__redis = redis.Redis(host, port, db)
        self.__expiration_time = expiration_time

    def __make_key(self, key: bytes, suffix: str) -> str:
        return f'liara/{key.hex()}/{suffix}'

    def _put(self, key: bytes, value: object) -> bool:
        if isinstance(value, bytes) or isinstance(value, bytearray):
            object_type = 'bin'
        else:
            object_type = 'obj'
            value = pickle.dumps(value)

        pipeline = self.__redis.pipeline()
        pipeline.set(self.__make_key(key, 'content'),
                     value, ex=self.__expiration_time)
        pipeline.set(self.__make_key(key, 'type'),
                     object_type, ex=self.__expiration_time)

        return all(pipeline.execute())

    def _get(self, key) -> Optional[object]:
        pipeline = self.__redis.pipeline()

        pipeline.get(self.__make_key(key, 'type'))
        pipeline.get(self.__make_key(key, 'content'))

        object_type, value = pipeline.execute()

        # Can't continue if any of those is None: Without the type we don't
        # know what to decode, and without a value the result is None anyways
        if object_type is None or value is None:
            return None

        if object_type == b'obj':
            return pickle.loads(value)

        return value

    def clear(self):
        for key in self.__redis.scan_iter('liara/*'):
            self.__redis.delete(key)

    def inspect(self):
        return CacheInfo(name='Redis')


class NullCache(_CacheBase):
    """The null cache drops all requests and does not cache any data.

    This is mostly useful to disable caching in APIs which require a cache
    instance.
    """
    def __init__(self):
        super().__init__()

    def _put(self, key: bytes, value: object) -> bool:
        return True

    def _get(self, key: bytes) -> Optional[object]:
        return None

    def clear(self) -> None:
        pass

    def inspect(self) -> CacheInfo:
        return CacheInfo(0, 0, 'Null')



from .nodes import GeneratedNode, NodeKind, DocumentNode
from .site import Site
from . import __version__
from .util import local_now
import datetime
import email.utils


class FeedNode(GeneratedNode):
    def __init__(self, path):
        super().__init__(path)


class RSSFeedNode(FeedNode):
    """A `RSS 2.0`_ based feed.
    
    .. _RSS 2.0: http://www.rssboard.org/rss-specification"""
    def __init__(self, path, site: Site,
                 *, collection='', limit=10):
        super().__init__(path)
        self.__collection = collection
        self.__limit = limit
        self.__site = site

    def generate(self):
        from lxml.builder import ElementMaker
        from lxml import etree

        items = reversed(list(self.__site.get_collection(
            self.__collection).nodes)[-self.__limit:])

        meta = self.__site.metadata

        E = ElementMaker(nsmap={
            'atom': "http://www.w3.org/2005/Atom"
        })
        A = ElementMaker(namespace='http://www.w3.org/2005/Atom')
        r = E.rss(version='2.0')

        c = E.channel(
            # See: http://www.rssboard.org/rss-profile#namespace-elements-atom
            A.link(href=meta['base_url'] + str(self.path), rel='self',
                   type='application/rss+xml'),
            E.title(meta['title']),
            E.link(meta['base_url']),
            E.description(meta['description']),
            E.generator(f'Liara {__version__}'),
            E.language(meta['language']),
            E.copyright(meta['copyright']),
            E.lastBuildDate(email.utils.format_datetime(
                local_now())),
        )

        for item in items:
            assert isinstance(item, DocumentNode)
            e = E.item(
                E.title(item.metadata['title']),
                E.link(meta['base_url'] + str(item.path)),
                E.pubDate(email.utils.format_datetime(item.metadata['date'])),
                E.guid(meta['base_url'] + str(item.path)),
                E.description(item.content)
            )
            c.append(e)
        r.append(c)

        self.content = etree.tostring(r)


class JsonFeedNode(FeedNode):
    """A `JSONFeed`_ based feed.
    
    .. _JSONFeed: https://jsonfeed.org"""
    def __init__(self, path, site: Site,
                 *, collection='', limit=10):
        super().__init__(path)
        self.__collection = collection
        self.__limit = limit
        self.__site = site

    def generate(self):
        import json
        items = reversed(list(self.__site.get_collection(
            self.__collection).nodes)[-self.__limit:])

        meta = self.__site.metadata

        result = {
            'version': 'https://jsonfeed.org/version/1',
            'title': meta['title'],
            'home_page_url': meta['base_url'],
            'feed_url': meta['base_url'] + str(self.path),
            'description': meta['description']
        }

        result_items = []
        for item in items:
            assert isinstance(item, DocumentNode)
            result_items.append({
                'id': meta['base_url'] + str(item.path),
                'title': item.metadata['title'],
                'date_published': item.metadata['date'].isoformat('T'),
                'url': meta['base_url'] + str(item.path),
                'content_html': item.content
            })

        result['items'] = result_items
        self.content = json.dumps(result)


class SitemapXmlFeedNode(FeedNode):
    """A `Sitemap 0.90`_ based feed.
    
    .. _Sitemap 0.90: https://www.sitemaps.org/"""
    def __init__(self, path, site: Site):
        super().__init__(path)
        self.__site = site

    def generate(self):
        from lxml.builder import ElementMaker
        from lxml import etree

        E = ElementMaker(
            namespace='http://www.sitemaps.org/schemas/sitemap/0.9',
            nsmap={
                None: 'http://www.sitemaps.org/schemas/sitemap/0.9'
            })

        metadata = self.__site.metadata
        now = datetime.datetime.now()

        urlset = E.urlset()
        for node in self.__site.nodes:
            if node.kind not in {NodeKind.Document, NodeKind.Index}:
                continue

            url = E.url()
            url.append(E.loc(metadata['base_url'] + str(node.path)))

            if 'date' in node.metadata:
                url.append(E.lastmod(node.metadata['date'].isoformat()))
            else:
                url.append(E.lastmod(now.isoformat()))

            if node.kind == NodeKind.Index:
                # This is machine generated, so reduce priority
                url.append(E.priority('0'))
            urlset.append(url)
        self.content = etree.tostring(urlset)



import pathlib
import collections.abc
import datetime
import tzlocal
from typing import List, Optional
import os
import fnmatch


def pairwise(iterable):
    """For a list ``s``, return pairs for consecutive entries. For example,
    a list ``s0``, ``s1``, etc. will produce ``(s0,s1), (s1,s2), ...`` and so
    on.

    See: https://docs.python.org/3/library/itertools.html#recipes."""
    import itertools
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def add_suffix(path: pathlib.PurePosixPath, suffix):
    """Add a suffix to a path.

    This differs from ``with_suffix`` by adding a suffix without changing
    the extension, i.e. adding ``en`` to ``foo.baz`` will produce
    ``foo.en.baz``."""
    name = path.name
    name, _, extension = name.rpartition('.')
    return path.with_name(name + '.' + suffix + '.' + extension)


def readtime(wordcount: int, words_per_minute=300):
    """Given a number of words, estimate the time it would take to read
    them.

    :return: The time in minutes if it's more than 1, otherwise 1."""
    return max(1, round(wordcount / words_per_minute))


def flatten_dictionary(d, sep='.', parent_key=None,
                       *, ignore_keys: Optional[set] = None):
    """Flatten a nested dictionary. This uses the separator to combine keys
    together, so a dictionary access like ``['a']['b']`` with a separator
    ``'.'`` turns into ``'a.b'``.

    If ``ignore_keys`` is set, it must be a list of fully flattened key names
    at which the flattening should stop. For instance, if a dictionary
    ``{'a': {'b': {'c': 1}}}`` is provided, and ``ignore_keys`` is ``{'a.b'}``,
    then ``a.b`` will not get flattened further, so ``a.b`` will contain a
    dictionary with ``{'c': 1}``.
    """
    items = []
    ignore_keys = ignore_keys if ignore_keys else set()

    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.abc.Mapping) \
           and new_key not in ignore_keys:
            items.extend(flatten_dictionary(v, sep=sep,
                                            parent_key=new_key,
                                            ignore_keys=ignore_keys).items())
        else:
            items.append((new_key, v,))
    return dict(items)


def create_slug(s: str) -> str:
    """Convert a plain string into a slug.

    A slug is suitable for use as a URL. For instance, passing ``A new world``
    to this function will return ``a-new-world``.
    """
    import slugify
    return slugify.slugify(s)


__TZ = tzlocal.get_localzone()
__override_now = None


def set_local_now(dt: datetime.datetime):
    """
    Override "now" to allow previewing the page at a different point in time.
    """
    global __override_now
    __override_now = dt.astimezone(__TZ)


def local_now() -> datetime.datetime:
    """Get the current date/time in the local time zone.

    This is equivalent to ``datetime.datetime.now()``, except it returns a
    timestamp which has ``tzinfo`` set to the local timezone.

    This can be overridden using ``set_local_now`` to build the page at a
    different point in time.
    """
    if __override_now:
        return __override_now
    return datetime.datetime.now(tz=__TZ)


class FilesystemWalker:
    def __init__(self, ignore_files: Optional[List[str]] = None):
        self.__ignore_files = ignore_files if ignore_files else []

    def walk(self, path: pathlib.Path):
        """Walk a directory recursively.

        This is quite similar to ``os.walk``, but with two major differences:

        * Files matching the ``ignore_files`` pattern are ignored.
        * The ``dirnames`` part of the tuple is omitted
        """
        for dirpath, _, filenames in os.walk(path):
            files_to_ignore = set()
            for pattern in self.__ignore_files:
                for filename in fnmatch.filter(filenames, pattern):
                    files_to_ignore.add(filename)

            filenames = [filename for filename in filenames
                         if filename not in files_to_ignore]

            yield dirpath, filenames


def get_hash_key_for_map(m: collections.abc.Mapping) -> bytes:
    """Calculate a hash key for any kind of map, including chain maps.
    
    This method is fairly slow. Use with caution."""
    import hashlib
    import pickle
    import io
    buffer = io.BytesIO()
    pickle.dump(m, buffer)
    return hashlib.sha1(buffer.getbuffer()).digest()

def file_digest(fp) -> bytes:
    """Back-compat implementation of Python's 3.11 `file_digest`"""
    import hashlib

    if hasattr(hashlib, 'file_digest'):
        return hashlib.file_digest(fp, 'sha512').digest()

    hasher = hashlib.new('sha512')
    while buffer := fp.read(1024 * 1024):
        hasher.update(buffer)
    return hasher.digest()



from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor
from markdown.preprocessors import Preprocessor
import re
import logging
from typing import (
    Any, Dict, Optional, Sequence
)

class HeadingLevelFixupProcessor(Treeprocessor):
    """This processor demotes headings by one level.

    By default, Markdown starts headings with ``<h1>``, but in general the
    title will be provided by a template. This processor replaces each heading
    with the next-lower heading, and adds a ``demoted`` class.
    """
    def run(self, root):
        return self._demote_header(root)

    def _demote_header(self, element):
        _demotions = {
            'h1': 'h2',
            'h2': 'h3',
            'h3': 'h4',
            'h4': 'h5',
            'h5': 'h6',
            'h6': 'h6',
        }

        if demotion := _demotions.get(element.tag):
            element.tag = demotion
            element.set('class', 'demoted')

        for e in element:
            self._demote_header(e)


class ShortcodeException(Exception):
    def __init__(self, error_message, line_number,
                 line_offset: Optional[int] = None):
        super().__init__(self._format_message(error_message, line_number,
                                              line_offset))
        self.__error_message = error_message
        self.__line_number = line_number

    def with_line_offset(self, offset):
        return ShortcodeException(
            self.__error_message,
            self.__line_number,
            offset
        )

    def _format_message(self, error_message, line_number, line_offset):
        if line_offset is not None:
            return f'Line {line_number + line_offset}: {error_message}'
        else:
            return error_message


class _ParseBuffer:
    def __init__(self, first_line: int, lines: Sequence[str]):
        self.__current_line = first_line
        self.__lines = lines
        self.__line_start = 0

    def get_current_line(self) -> str:
        return self.__lines[self.__current_line][self.__line_start:]

    def get_current_line_number(self) -> int:
        return self.__current_line + 1

    def advance_line(self):
        self.__current_line += 1
        self.__line_start = 0

        if self.__current_line >= len(self.__lines):
            raise ShortcodeException(
                "Unexpected end of shortcode",
                self.__current_line)

    def advance(self, offset: int):
        self.__line_start += offset

    def peek(self):
        try:
            return self.__lines[self.__current_line][self.__line_start]
        except IndexError:
            return None

    def consume(self, what: str):
        """Consume ``what`` from the input, and raise an error otherwise.

        ``what`` must not contain a newline."""
        assert '\n' not in what
        line = self.get_current_line()
        if line.startswith(what):
            self.advance(len(what))
        else:
            raise ShortcodeException(
                f'Shortcode parse error: Expected "{what}"',
                self.get_current_line_number())


class _ShortcodeParser:
    def __init__(self, buffer: _ParseBuffer):
        self.__args = dict()
        self.__function_name = None
        self.__buffer = buffer
        self.__key_re = re.compile(r'^\w+')
        self.__arg_re = re.compile(r'^[\w-]+')
        self.__whitespace_re = re.compile(r'\s')

    def parse(self):
        """Parse a shortcode starting at the current position in the
        buffer.

        Returns a tuple containing: Any text following the short code in the
        last line, the next line to resume parsing from, the function name,
        and the arguments."""
        self.__consume('<%')
        self.__consume_whitespace()
        self.__function_name = self.__consume_key('function name')
        self.__consume_whitespace()

        while self.__buffer.peek() != '/':
            key = self.__consume_key('argument name')

            self.__consume_whitespace()
            self.__consume('=')
            self.__consume_whitespace()

            if self.__buffer.peek() == '"':
                value = self.__consume_string()
            else:
                value = self.__consume_arg()

            self.__args[key] = value
            self.__consume_whitespace()

        self.__consume('/%>')

        return (self.__buffer.get_current_line(),
                # This is actually the line number + 1, but that's exactly the
                # line at which parsing should continue, so this works out
                self.__buffer.get_current_line_number(),
                self.__function_name,
                self.__args,)

    def __consume(self, what):
        self.__buffer.consume(what)

    def __consume_arg(self) -> str:
        """Consume a function argument.

        Works on the current line only, may advance to the end of the line.
        """
        # args cannot span lines
        # an argument is anything that doesn't end in /%>, but we also disallow
        # whitespace and newlines. We thus search for the first whitespace
        # character or ``/%>``
        line = self.__buffer.get_current_line()
        if match := self.__arg_re.search(line):
            arg = line[:match.end()]
            self.__buffer.advance(match.end())

            return arg
        else:
            raise ShortcodeException(
                    'Error while parsing shortcode: Could not parameter value',
                    self.__buffer.get_current_line_number()
            )

    def __consume_whitespace(self):
        """Consume whitespace from the buffer.

        This will consume multiple lines if needed.
        """
        while True:
            line = self.__buffer.get_current_line()
            if not line:
                self.__buffer.advance_line()
                continue

            if match := self.__whitespace_re.search(line):
                if match.start() == 0:
                    self.__buffer.advance(match.end())
                else:
                    return
            else:
                return

    def __consume_string(self) -> str:
        """Consume a quoted string from the buffer.

        This will consume multiple lines if needed.
        """
        # Strings can be multi-line while inside "
        # Our strategy is as following: Find the next ", check if it was
        # escaped or not, and continue until we find one
        self.__buffer.consume('"')
        result = []
        while True:
            line = self.__buffer.get_current_line()
            next_quote = 0
            while True:
                if (quote_position := line.find('"', next_quote)) != -1:
                    if quote_position > 0 and line[quote_position-1] != '\\':
                        result.append(line[:quote_position])
                        self.__buffer.advance(quote_position)
                        self.__buffer.consume('"')
                        return ''.join(result).replace('\\"', '"')
                    else:
                        next_quote = quote_position+1
                else:
                    # No terminating quote in this line
                    result.append(line)
                    self.__buffer.advance_line()
                    break

    def __consume_key(self, what: str) -> str:
        """Consume a key from the buffer.

        Works on the current line only, may advance to the end of the line.
        """
        # keys cannot span lines, so we get the current line from the buffer
        # and match from the current position on
        line = self.__buffer.get_current_line()
        if match := self.__key_re.search(line):
            assert match.start() == 0
            key = line[:match.end()]
            self.__buffer.advance(match.end())

            return key
        else:
            raise ShortcodeException(
                    f'Error while parsing shortcode: Could not parse {what}',
                    self.__buffer.get_current_line_number()
                    )


class ShortcodePreprocessor(Preprocessor):
    """
    A Wordpress-inspired "shortcode" preprocessor which allows calling
    functions before the markup processing starts.

    Shortcodes are delimited by ``<%`` and ``/%>``. The content must start with
    the function name, followed by ``key=value`` pairs. The values are passed
    as strings, the calling function must do the type conversion.

    Values without quotation marks must consist of alphanumeric characters and
    ``-``, ``_`` only.

    .. note::
      Freestanding keys are not supported as it's ambiguous whether this should
      be a considered a plain argument (i.e. passed as a non-named argument) or
      a defaulted argument (i.e. named, but set to ``True`` or some other
      value.) For example, you can't write a shortcode like this::

        <% alert message="This is important" blink /%>

      Instead, you'll have to use something like::

        <% alert message="This is important" blink=yes /%>
    """

    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, md=None, node=None):
        from .template import Page
        super().__init__(md)
        self.__functions = dict()
        self.__page = Page(node)
        self.__data = None

    def set_data(self, data: Dict[str, Any]):
        """
        Set the data context.

        .. versionadded:: 2.6.2
        """
        self.__data = data

    def register(self, name: str, function):
        """Register a new Markdown shortcode function.

        Shortcode function calls must accept all arguments as named arguments.
        Names (both function names and argument names) starting with ``$`` are
        reserved for built-in functions.

        Shortcode handlers must accept a final ``**kwargs`` argument to handle
        any context Liara may pass in. Liara context variables will be prefixed
        with ``$`` which is disallowed as an parameter name otherwise.
        """
        import inspect

        signature = inspect.signature(function)
        if not any([p.kind == p.VAR_KEYWORD
                    for p in signature.parameters.values()]):
            raise Exception(f'Cannot register function "{name}" as a '
                            'shortcode handler as the function signature '
                            'is missing a **kwargs parameter.')

        assert name and name[0] != '$'
        self.__functions[name] = function

    def run(self, lines: list[str]):
        i = 0
        line_count = len(lines)

        # We can't use for ... in range or for ... enumerate here, as we have
        # to skip lines based on the shortcode parser
        while i < line_count:
            line = lines[i]

            if (tag_start := line.find('<%')) != -1:
                # Only yield if the string is non-empty. Same logic below for
                # rest
                if start := line[:tag_start]:
                    yield start

                # Remove start from the current line
                lines[i] = line[tag_start:]

                parse_buffer = _ParseBuffer(i, lines)
                shortcode_parser = _ShortcodeParser(parse_buffer)
                rest, next_line, func_name, args = shortcode_parser.parse()

                args['$page'] = self.__page
                if self.__data:
                    args['$data'] = self.__data

                def pretty_print_args(d):
                    for k, v in d.items():
                        # Skip internal arguments as they can't be
                        # pretty-printed anyways
                        if k[0] == '$':
                            continue
                        yield f'{k}={repr(v)}'

                # Skip the text processing if logging is disabled
                if self.__log.isEnabledFor(logging.DEBUG):
                    self.__log.debug('Calling shortcode handler: "%s" with '
                                     'arguments: %s',
                                     func_name,
                                     ', '.join(pretty_print_args(args)))
                yield from self.__functions[func_name](**args).splitlines()

                # Another shortcode in the same line, so we need to resume
                # parsing there and cannot simply emit the rest
                if '<%' in rest:
                    assert next_line >= 1
                    lines[next_line-1] = rest
                    continue

                if rest:
                    yield rest

                i = next_line
            else:
                yield line
                i += 1


class LiaraMarkdownExtensions(Extension):
    """Register various markdown extensions.
    """
    def __init__(self, node=None):
        super().__init__()
        self.__node = node
        self.__shortcode_preprocessor = None

    def set_data(self, data: Dict[str, Any]):
        assert self.__shortcode_preprocessor
        self.__shortcode_preprocessor.set_data(data)

    def extendMarkdown(self, md):
        from .signals import register_markdown_shortcodes

        self.__shortcode_preprocessor = ShortcodePreprocessor(md, self.__node)

        register_markdown_shortcodes.send(
            self,
            preprocessor=self.__shortcode_preprocessor)

        md.treeprocessors.register(HeadingLevelFixupProcessor(md),
                                   'heading-level-fixup',
                                   100)
        md.preprocessors.register(self.__shortcode_preprocessor,
                                  'shortcode-preprocessor',
                                  100)
        md.registerExtension(self)



from .nodes import (
    DataNode,
    DocumentNode,
    GeneratedNode,
    IndexNode,
    Node,
    NodeKind,
    ResourceNode,
    StaticNode,
    ThumbnailNode,
)
import pathlib
from typing import (
    Any,
    Dict,
    Iterable,
    KeysView,
    List,
    Optional,
    Tuple,
    Union,
    ValuesView,
)
from .util import pairwise
from . import signals
import logging
import fnmatch
from abc import abstractmethod, ABC
from types import MappingProxyType


def _create_metadata_accessor(field_name):
    """Create an accessor for nested metadata fields.

    This allows accessing attributes or dictionary entries recursively, i.e.
    an accessor ``foo.bar.baz`` will access ``foo['bar']['baz]`` and
    ``foo.bar.baz``, and any combination thereof. Attributes will be preferred
    over dictionary access.

    If the field is not present, this function returns ``None``.
    """
    if '.' in field_name:
        def dict_key_fun(o):
            field = field_name.split('.')
            o = o.metadata.get(field[0])
            if o is None:
                return None
            for f in field[1:]:
                if hasattr(o, f):
                    o = getattr(o, f)
                elif f in o:
                    o = o[f]
                else:
                    return None
            return o
        return dict_key_fun
    else:
        def metadata_key_fun(o):
            return o.metadata.get(field_name)
        return metadata_key_fun


def _create_metadata_filter(exclude_without:
                            List[Union[str, Tuple[str, Any]]]):
    """Create a function to filter out nodes based on metadata fields.

    The filter function will check if the specified fields are present,
    and, if a tuple has been passed, if that field matches the expected
    value."""
    key_functions = []
    for f in exclude_without:
        if isinstance(f, str):
            key = f

            def key_function(node: Node):
                return _create_metadata_accessor(key)(node) is not None
            key_functions.append(key_function)
        else:
            assert isinstance(f, tuple)
            assert len(f) == 2
            key = f[0]

            def key_function(node: Node):
                return _create_metadata_accessor(key)(node) == f[1]
            key_functions.append(key_function)

    def filter_function(node: Node):
        for key_function in key_functions:
            if not key_function(node):
                return False
        return True

    return filter_function


class Collection:
    """A collection is a set of nodes. Collections can be ordered, which allows
    for next/previous queries.
    """

    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, site: 'Site', name: str, pattern: str, *,
                 exclude_without: Optional[List[Union[str, Tuple[str, Any]]]]
                 = None,
                 order_by: Optional[Union[str, List[str]]] = None,
                 node_kinds: Optional[List[Union[str, NodeKind]]] = None):
        """
        Create a new collection.

        :param pattern: The pattern to select nodes which belong to this
                        collection.
        :param exclude_without: Exclude items without the specified metadata
                          fields. If a tuple is provided, the metadata field's
                          value must match the requested value.
        :param order_by: A list of accessors for fields to order by. If
                         multiple entries are provided, the result will be
                         sorted by each in order using a stable sort. To
                         reverse the order, use a leading ``-``, for
                         example: ``-date``.
        :param node_kinds: Only include nodes of that kinds. If not specified,
                           ``NodeKind.Document`` will be used as the default.

        If an ordering is specified, and a particular node cannot support that
        ordering (for instance, as it's missing the field that is used to order
        by), an error will be raised.
        """
        from .nodes import _parse_node_kind

        self.__name = name

        self.__site = site
        self.__filter = pattern
        self.__node_kinds = set()
        if node_kinds:
            for kind in node_kinds:
                if isinstance(kind, NodeKind):
                    self.__node_kinds.add(kind)
                else:
                    self.__node_kinds.add(_parse_node_kind(kind))
        else:
            self.__node_kinds.add(NodeKind.Document)

        self.__order_by = order_by if order_by else []

        self.__exclude_without = exclude_without
        self.__build()

    def __build(self):
        nodes = self.__site.select(self.__filter)

        def filter_kind(node):
            return node.kind in self.__node_kinds

        nodes = filter(filter_kind, nodes)

        if self.__exclude_without:
            nodes = filter(_create_metadata_filter(self.__exclude_without),
                           nodes)

        for ordering in self.__order_by:
            if ordering.startswith('-'):
                ordering = ordering[1:]
                reverse = True
            else:
                reverse = False

            def key_fun(node):
                accessor = _create_metadata_accessor(ordering)
                result = accessor(node)
                if result is None:
                    self.__log.error(
                        'Node "%s" is missing the metadata '
                        'field "%s" which is required by a '
                        'order_by statement. Use exclude_without to '
                        'excludes nodes which miss certain metadata '
                        'fields.',
                        node.path, ordering)
                    raise Exception(f'Cannot order collection "{self.__name}" '
                                    f'by key "{ordering}"')
                return result

            nodes = sorted(nodes, key=key_fun, reverse=reverse)

        # Need to convert to a list here, as we're going to iterate over it
        # twice (once for the pairs, once for the nodes property)
        nodes = list(nodes)

        pairs = list(pairwise(nodes))
        self.__next = {i[0].path: i[1] for i in pairs}
        self.__previous = {i[1].path: i[0] for i in pairs}
        self.__nodes = {n.path: n for n in nodes}

    @property
    def nodes(self) -> Iterable[Node]:
        """Get the (sorted) nodes in this collection."""
        return self.__nodes.values()

    def get_next(self, node: Node) -> Optional[Node]:
        """Get the next node in this collection with regard to the specified
        order, or ``None`` if this is the last node."""
        if not self.__order_by:
            self.__log.warning(
                'Using get_next() on an unordered collection has '
                'undefined results')
        return self.__next.get(node.path)

    def get_previous(self, node: Node) -> Optional[Node]:
        """Get the previous node in this collection with regard to the
        specified order, or ``None`` if this is the first node."""
        if not self.__order_by:
            self.__log.warning(
                'Using get_next() on an unordered collection has '
                'undefined results')
        return self.__previous.get(node.path)


def _group_recursive(iterable, group_keys: List[str]):
    import collections
    import itertools
    if not group_keys:
        return list(iterable)

    current_key = group_keys[0]
    group_keys = group_keys[1:]
    if current_key[0] == '*':
        key_func = _create_metadata_accessor(current_key[1:])
        # If the group key starts with a *, this means it's a list of items,
        # not a single item to group by -- for instance, a list of tags
        # In this case, we create one bucket per item, append the items to that
        # list, and eventually we recurse if needed
        result = collections.defaultdict(list)
        for item in iterable:
            groups = key_func(item)
            for group in groups:
                result[group].append(item)

        if group_keys:
            return {k: _group_recursive(v, group_keys)
                    for k, v in result.items()}
        return result
    else:
        key_func = _create_metadata_accessor(current_key)
        iterable = sorted(iterable, key=key_func)
        result = dict()
        for k, v in itertools.groupby(iterable, key_func):
            values = _group_recursive(v, group_keys)
            result[k] = values
        return result


class Index:
    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    """An index into a collection, which provides an index structure.

    The index structure requires a grouping schema -- for instance, all nodes
    containing some tag can get grouped under one index node.
    """
    def __init__(self, collection: Collection,
                 path: str, group_by: List[str], *,
                 exclude_without: Optional[List[Union[str, Tuple[str, Any]]]]
                 = None,
                 create_top_level_index=False):
        """
        Create a new index.

        :param collection: The collection to use for this index.
        :param path: The path pattern to create index entries at. This must
                     contain one numbered placeholder (``%1``, etc.) per entry
                     in ``group_by``
        :param group_by: The grouping statements. The nodes are grouped by each
                         entry in this array in-order.
        :param exclude_without: Exclude items without the specified metadata
                          field. If a tuple is provided, the metadata field's
                          value must match the requested value.
        :param create_top_level_index: Create a node at the top-level path as
                                       well instead of only creating nodes
                                       per grouping statement.
        """
        nodes = collection.nodes

        if exclude_without:
            filter_function = _create_metadata_filter(exclude_without)
            nodes = filter(filter_function, nodes)

        assert len(group_by) > 0
        self.__groups = _group_recursive(nodes, group_by)
        self.__path = path
        self.__create_top_level_index = create_top_level_index
        self.__top_level_node: IndexNode | None = None

    def create_nodes(self, site: 'Site'):
        """Create the index nodes inside the specified site."""
        if self.__create_top_level_index:
            url = self.__path
            # Strip off components until we have no parameter left
            while '%' in url:
                url = url[:url.rfind('/')]
            node = IndexNode(pathlib.PurePosixPath(url), {
                'path': self.__path,
                'top_level_index': True
            })
            site.add_index(node)
            self.__top_level_node = node
            self.__log.debug('Created top-level index at "%s"',
                             self.__path)

        self._create_nodes_recursive(site, self.__path,
                                     self.__groups, 1,
                                     self.__top_level_node)

    def _create_nodes_recursive(self, site: 'Site', path, d, index,
                                parent: Optional[IndexNode] = None):
        for k, v in d.items():
            url = pathlib.PurePosixPath(path.replace(f'%{index}', str(k)))
            # TODO Find out what to do here -- either don't create intermediate
            # nodes (i.e. as long as one %N placeholder is left), or cut off
            # the URL.
            node = IndexNode(url, {'key': k})
            site.add_index(node)

            if parent:
                parent.add_reference(node)

            if isinstance(v, dict):
                self._create_nodes_recursive(site, url, v, index+1, node)
            else:
                for reference in v:
                    node.add_reference(reference)


class ContentFilter(ABC):
    """Content filters can filter out nodes based on various criteria."""
    @abstractmethod
    def apply(self, node: Node) -> bool:
        """Return ``True`` if the node should be kept, and ``False``
        otherwise.
        """
        ...

    @property
    def reason(self) -> str:
        """Return a reason why this filter applied."""
        return ''

    @property
    def name(self) -> str:
        return self.__class__.__name__


class DateFilter(ContentFilter):
    """Filter content based on the metadata field ``date``.

    If the date is in the future, the node will be filtered."""
    def __init__(self):
        from .util import local_now
        self.__now = local_now()

    def apply(self, node: Node) -> bool:
        date = node.metadata.get('date', None)
        if date is None:
            return True

        return date <= self.__now

    @property
    def reason(self):
        return f'`date` <= {self.__now}'

    @property
    def name(self):
        return 'date'


class StatusFilter(ContentFilter):
    """Filter content based on the metadata field ``status``.

    If ``status`` is set to ``private``, the node will be filtered. The
    comparison is case-insensitive.
    """
    def apply(self, node: Node) -> bool:
        return node.metadata.get('status', '').lower() != 'private'

    @property
    def reason(self):
        return '`status` set to "private"'

    @property
    def name(self):
        return 'status'


class ContentFilterFactory:
    def __init__(self):
        self.__filters = {
            'date': DateFilter,
            'status': StatusFilter
        }

    def create_filter(self, name: str) -> ContentFilter:
        class_ = self.__filters[name]
        return class_()


class Site:
    """This class manages to all site content."""

    data: List[DataNode]
    """The list of all data nodes in this site."""

    indices: List[IndexNode]
    """The list of all index nodes in this site."""

    documents: List[DocumentNode]
    """The list of all document nodes in this site."""

    resources: List[ResourceNode]
    """The list of all resources nodes in this site."""

    static: List[StaticNode]
    """The list of all static nodes in this site."""

    generated: List[GeneratedNode]
    """The list of all generated nodes in this site."""

    metadata: Dict[str, Any]
    """Metadata describing this site."""

    __nodes: Dict[pathlib.PurePosixPath, Node]
    __root = pathlib.PurePosixPath('/')
    __collections: Dict[str, Collection]
    __indices: List[Index]
    __content_filters: List[ContentFilter]
    __filtered_content: Dict[pathlib.PurePosixPath, str]

    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    __merged_data: Dict[str, Any]

    def __init__(self):
        self.data = []
        self.indices = []
        self.documents = []
        self.resources = []
        self.static = []
        self.generated = []
        self.metadata = {}
        self.__nodes = {}
        self.__collections = {}
        self.__indices = []
        self.__content_filters = []

        # Stores the paths of filtered nodes, and the filter that filtered them
        self.__filtered_content = {}

        # Store the union of the data provided in data nodes
        self.__merged_data = {}

    def register_content_filter(self, content_filter: ContentFilter):
        """Register a new content filter."""
        self.__content_filters.append(content_filter)

    def set_metadata(self, metadata: Dict[str, Any]) -> None:
        """Set the metadata for this site.

        This overrides any previously set metadata. Metadata is accessible
        via the :py:attr:`metadata` attribute.
        """
        self.metadata = metadata

    def set_metadata_item(self, key: str, value: Any) -> None:
        """Set a single entry in the metadata for this site.

        This can be used to override individual metadata items.
        """
        self.metadata[key] = value

    def __is_content_filtered(self, node: Node) -> bool:
        for f in self.__content_filters:
            if not f.apply(node):
                self.__log.info('Filtered node "%s" due to reason: %s',
                                node.path, f.reason)
                self.__filtered_content[node.path] = f.name
                signals.content_filtered.send(self, node=node, filter=f)
                return True
        return False

    @property
    def filtered_content(self) -> Dict[pathlib.PurePosixPath, str]:
        """Return which node paths got filtered and due to which filter."""
        return self.__filtered_content

    def add_data(self, node: DataNode) -> None:
        "Add a data node to this site."""
        self.data.append(node)
        self.__register_node(node)

        self.__merged_data.update(node.content)

    def add_index(self, node: IndexNode) -> None:
        """Add an index node to this site."""
        self.indices.append(node)
        self.__register_node(node)

    def add_document(self, node: DocumentNode) -> None:
        """Add a document to this site."""
        if self.__is_content_filtered(node):
            return

        self.documents.append(node)
        self.__register_node(node)

    def add_resource(self, node: ResourceNode) -> None:
        """Add a resource to this site."""
        if self.__is_content_filtered(node):
            return

        self.resources.append(node)
        self.__register_node(node)

    def add_static(self, node: StaticNode) -> None:
        """Add a static node to this site."""
        if self.__is_content_filtered(node):
            return

        self.static.append(node)
        self.__register_node(node)

    def add_generated(self, node: GeneratedNode) -> None:
        "Add a generated node to this site."""
        self.generated.append(node)
        self.__register_node(node)

    def __register_node(self, node: Node) -> None:
        if node.path in self.__nodes:
            raise Exception(f'"{node.path}" already exists, '
                            'cannot overwrite.')

        signals.content_added.send(self, node=node)
        self.__nodes[node.path] = node

    @property
    def nodes(self) -> ValuesView[Node]:
        """The list of all nodes in this site."""
        return self.__nodes.values()

    @property
    def urls(self) -> KeysView[pathlib.PurePosixPath]:
        """The list of all registered URLs."""
        return self.__nodes.keys()

    def create_links(self):
        """This creates links between parents/children.

        This is a separate step so it can be executed after merging nodes
        from multiple sources, for instance themes. It is safe to call this
        function multiple times to create new links; nodes which already
        have a parent are automatically skipped."""
        for key, node in self.__nodes.items():
            if node.parent:
                # The parent is already set, so there's nothing to do
                continue

            parent_path = key.parent
            parent_node = self.__nodes.get(parent_path)
            # The parent_node != node check is required so the root node
            # doesn't get added to itself (by virtue of / being a parent of /)
            if parent_node and parent_node != node:
                parent_node.add_child(node)

    def create_collections(self, collections):
        """Create collections."""
        for name, collection in collections.items():
            order_by = collection.get('order_by', [])
            if isinstance(order_by, str):
                order_by = [order_by]

            exclude_without = collection.get('exclude_without', [])
            if isinstance(exclude_without, str):
                exclude_without = [exclude_without]

            node_kinds = collection.get('node_kinds', [])

            self.__log.debug('Creating collection "%s" ... ', name)
            collection = Collection(self, name,
                                    collection['filter'],
                                    order_by=order_by,
                                    exclude_without=exclude_without,
                                    node_kinds=node_kinds)
            self.__collections[name] = collection
            self.__log.debug('... done creating collection "%s" ... ', name)

    def create_indices(self, indices):
        """Create indices."""
        for index_definition in indices:
            collection = self.get_collection(index_definition['collection'])
            del index_definition['collection']
            create_top_level_index = index_definition.get(
                'create_top_level_index', False
            )

            group_by = index_definition.get('group_by')
            if isinstance(group_by, str):
                group_by = [group_by]

            exclude_without = index_definition.get('exclude_without', [])
            if isinstance(exclude_without, str):
                exclude_without = [exclude_without]

            self.__log.debug('Creating index for path "%s" ...',
                             index_definition['path'])
            index = Index(collection, index_definition['path'],
                          group_by=group_by,
                          exclude_without=exclude_without,
                          create_top_level_index=create_top_level_index)
            self.__indices.append(index)
            self.__log.debug('... done creating index for path "%s"',
                             index_definition['path'])

        for index in self.__indices:
            index.create_nodes(self)

        # Indices may have been added, so we need to update the links
        self.create_links()

    def create_thumbnails(self, thumbnail_definition):
        """Create thumbnails.

        Based on the thumbnail definition -- which is assumed to be a
        dictionary containing the suffix, the desired size and the target
        formats  -- this function iterates over all static nodes that contain
        images, and creates new thumbnail nodes as required.
        """
        from .util import add_suffix

        def create_thumbnail(new_path, format, size):
            if format == 'original':
                format = None
            else:
                new_path = new_path.with_suffix(
                    f'.{format.lower()}'
                )

            if self.get_node(new_path):
                # This happens when the thumbnail format is the same as the
                # image format, and both 'original' and the format are
                # specified
                # This isn't really a problem and fixing this at a higher level
                # isn't any easier than doing it here
                self.__log.debug(
                    'Skipping ".%s" thumbnail creation for "%s" as '
                    'that image already exists as "%s"',
                    format if format else static.src.suffix[1:],
                    static.src,
                    new_path)
                return

            thumbnail = ThumbnailNode(
                static.src,
                new_path,
                size,
                format)
            self.add_resource(thumbnail)

        new_static = []
        for static in self.static:
            if not static.is_image:
                continue
            static.update_metadata()
            width, height = static.metadata['image_size']
            for k, v in thumbnail_definition['sizes'].items():
                if 'exclude' in v and 'include' in v:
                    self.__log.warning('Thumbnail size "%s" has both '
                                       '"include" and "exclude" filters '
                                       'enabled, "include" will be ignored.')

                if pattern := v.get('exclude'):
                    if fnmatch.fnmatch(static.src.name, pattern):
                        self.__log.debug(
                                'Skipping thumbnail creation for "%s" due to '
                                'exclude filter "%s"',
                                static.src, pattern)
                        continue
                elif pattern := v.get('include'):
                    if not fnmatch.fnmatch(static.src.name, pattern):
                        self.__log.debug(
                                'Skipping thumbnail creation for "%s" due to '
                                'include filter "%s"',
                                static.src, pattern)
                        continue

                new_url = add_suffix(static.path, k)
                thumbnail_width = v.get('width', width)
                thumbnail_height = v.get('height', height)
                if width <= thumbnail_width and height <= thumbnail_height:
                    # Check for any other formats present and generate those
                    # as needed
                    for format in thumbnail_definition['formats']:
                        if format == 'original':
                            # The image has the right size already (smaller or
                            # equal to the thumbnail), so we just link a new
                            # static node
                            self.__log.debug(
                                'Copying image "%s" for thumbnail as size '
                                '%d x %d is larger or equal than image size '
                                '%d x %d',
                                static.src,
                                width, height,
                                thumbnail_width, thumbnail_height)
                            copy = StaticNode(static.src, new_url)
                            copy.metadata = static.metadata
                            new_static.append(copy)
                        else:
                            self.__log.debug(
                                'Converting image "%s" for thumbnail due to '
                                'format change request to ".%s"',
                                static.src, format)
                            create_thumbnail(new_url, format, {})
                else:
                    for format in thumbnail_definition['formats']:
                        create_thumbnail(new_url, format, v)

        for static in new_static:
            self.add_static(static)

        # Static nodes may have been created, so we need to update the links
        self.create_links()

    def get_next_in_collection(self,
                               collection: str,
                               node: Node) -> Optional[Node]:
        """Get the next node in a collection."""
        next_node = self.__collections[collection].get_next(node)
        return next_node

    def get_previous_in_collection(self,
                                   collection: str,
                                   node: Node) -> Optional[Node]:
        """Get the previous node in a collection."""
        previous_node = self.__collections[collection].get_previous(node)
        return previous_node

    def get_collection(self, collection: str) -> Collection:
        """Get a collection."""
        return self.__collections[collection]

    def get_node(self, path: Union[str, pathlib.PurePosixPath]) \
            -> Optional[Node]:
        """Get a node based on the URL, or ``None`` if no such node exists."""
        if isinstance(path, str):
            path = pathlib.PurePosixPath(path)
        return self.__nodes.get(path)

    def select(self, query: str) -> Iterable[Node]:
        """Select nodes from this site.

        The query string may contain ``*`` to list all direct children of a
        node, and ``**`` to recursively enumerate nodes. Partial matches
        using ``*foo`` are not supported. See :doc:`/content/url-patterns` for
        details.
        """
        parts = query.split('/')
        # A valid query must start with /
        assert parts[0] == ''
        node = self.__nodes[self.__root]
        for component in parts[1:]:
            if component == '**':
                return node.get_children(recursive=True)
            if component == '*':
                return node.get_children(recursive=False)
            if component == '':
                return [node]

            if (next_child := node.get_child(component)) is None:
                return []
            else:
                node = next_child

        return [node]

    @property
    def merged_data(self) -> MappingProxyType[str, Any]:
        """Return the union of all data nodes.

        This is a read-only view, as shortcodes and templates can access the
        data in any order and thus modifications would be unsafe.
        
        .. versionadded:: 2.6.2
        """
        return MappingProxyType(self.__merged_data)


from enum import auto, Enum
import logging
import pathlib
from .yaml import load_yaml

try:
    import tomllib as toml
except ImportError:
    import tomli as toml

from .cache import Cache
from .signals import document_loaded
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
import re
import dateparser
from abc import abstractmethod, ABC

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .query import Query


class Publisher(ABC):
    """A publisher produces the final output files, applying templates etc. as
    needed.
    """
    @abstractmethod
    def publish_document(self, document: 'DocumentNode') -> pathlib.Path:
        """Publish a document node.

        :return: The path of the generated file.
        """
        ...

    @abstractmethod
    def publish_index(self, index: 'IndexNode') -> pathlib.Path:
        """Publish an index node.

        :return: The path of the generated file."""
        ...

    @abstractmethod
    def publish_resource(self, resource: 'ResourceNode') -> pathlib.Path:
        """Publish a resource node.

        :return: The path of the generated file."""
        ...

    @abstractmethod
    def publish_static(self, static: 'StaticNode') -> pathlib.Path:
        """Publish a static node.

        :return: The path of the generated file."""
        ...

    @abstractmethod
    def publish_generated(self, generated: 'GeneratedNode') -> pathlib.Path:
        """Publish a generated node.

        :return: The path of the generated file."""
        ...


class NodeKind(Enum):
    Resource = auto()
    Index = auto()
    Document = auto()
    Data = auto()
    Static = auto()
    Generated = auto()


class _AsyncTask(ABC):
    """A node can return an ``_AsyncTask`` from process, in which case the
    processing is split into two steps: The process itself (which can be
    executed on a separate process/in parallel), and a post-process step which
    can update the cache (in case the cache does not allow concurrent access.)

    The function must return the new content of the node.
    """
    @abstractmethod
    def process(self) -> object:
        """
        Process the task and return the new content for the node.
        """
        ...

    def update_cache(self, content: object, cache: Cache):
        """
        After processing, an async task may update the cache in a separate
        step.
        """
        pass


class Node(ABC):
    kind: NodeKind
    """The node kind, must be set in the constructor."""
    src: Optional[pathlib.Path]
    """The full path to the source file.

    This is an OS specific path object."""

    path: pathlib.PurePosixPath
    """The output path, relative to the page root.

    All paths *must* start with ``/``.
    """

    metadata: Dict[str, Any]
    """Metadata associated with this node."""

    parent: Optional['Node']
    """The parent node, if any."""

    __nodes: Dict[str, 'Node']
    """A dictionary containing all child nodes.

    The key is the path to the child node relative to this node. I.e. if the
    path of this node is ``/foo``, and it has a child at ``/foo/bar``, the
    key for that child would be ``bar``."""

    @property
    def children(self):
        """A list containing all direct children of this node."""
        return self.__nodes.values()

    def __init__(self):
        self.__nodes = {}
        self.metadata = {}
        self.parent = None

    def add_child(self, child: 'Node') -> None:
        """Add a new child to this node.

        The path of the child node must be a sub-path of the current node
        path, with exactly one more component. I.e. if the current node path is
        ``/foo/bar``, a node with path ``/foo/bar/baz`` can be added as a
        child, but ``/baz/`` or ``/foo/bar/boo/baz`` would be invalid."""
        assert self.path != child.path
        name = child.path.relative_to(self.path).parts[0]
        self.__nodes[name] = child
        child.parent = self

    def __repr__(self):
        return f'{self.__class__.__name__}({self.path})'

    def select_children(self) -> 'Query':
        """Select all children of this node and return them as a
        :py:class:`~liara.query.Query`."""
        from .query import Query
        return Query(self.children)

    def get_child(self, name) -> Optional['Node']:
        """Get a child of this node.

        :return: The child node or ``None`` if no such child exists."""
        return self.__nodes.get(name)

    def get_children(self, *, recursive=False) -> Iterable['Node']:
        """Get all children of this node.

        This function differs from :py:meth:`select_children` in two important
        ways:

        * It returns a list of :py:class:`Node` instances and does not wrap it
          in a :py:class:`~liara.query.Query`
        * It can enumerate all children recursively.
        """
        for child in self.children:
            yield child
            if recursive:
                yield from child.get_children(recursive=True)

    def process(self, cache: Cache, **kwargs) -> Optional[_AsyncTask]:
        """Some nodes -- resources, documents, etc. need to be processed. As
        this can be a resource-intense process (for instance, it may require
        generating images), processing can cache results and has to be
        called separately instead of being executed as part of some other
        operation.

        By convention this method should populate ``self.content`` if executed
        synchronously. If asynchronous execution is supported, this method
        must return an ``_AsyncTask`` which is then executed later. In this
        case, ``self.content`` will be populated by the task runner.

        The ``**kwargs`` are used to pass in additional context for processing.
        System-provided arguments will be prefixed with a `$` (which is not
        a valid identifier.)

        .. versionchanged:: 2.6.2 Added ``**kwargs``
           ``**kwargs`` was added to allow passing in additional context.
        """
        pass


_metadata_marker = re.compile(r'(---|\+\+\+)\n')


class MetadataKind(Enum):
    Unknown = auto()
    Yaml = auto()
    Toml = auto()


def extract_metadata_content(text: str):
    """Extract metadata and content.

    Metadata is stored at the beginning of the file, separated using a metadata
    separation marker, for instance::

      +++
      this_is_toml = True
      +++

      content

    This function splits the provided text into metadata and actual content.
    """
    meta_start, meta_end = 0, 0
    content_start, content_end = 0, 0
    metadata_kind = MetadataKind.Unknown

    # If the document doesn't end with a trailing new-line, the metadata regex
    # will get confused. We'll thus add a new-line to make sure this works
    if text and text[-1] != '\n':
        text += '\n'

    for match in _metadata_marker.finditer(text):
        if meta_start == 0:
            if match.group() == '---\n':
                metadata_kind = MetadataKind.Yaml
            elif match.group() == '+++\n':
                metadata_kind = MetadataKind.Toml
            meta_start = match.span()[1]
        elif meta_end == 0:
            if match.group() == '---\n':
                if metadata_kind != MetadataKind.Yaml:
                    raise Exception('Metadata markers mismatch -- started '
                                    'with "---", but ended with "+++"')
            elif match.group() == '+++\n':
                if metadata_kind != MetadataKind.Toml:
                    raise Exception('Metadata markers mismatch -- started '
                                    'with "+++", but ended with "---"')

            meta_end = match.span()[0]
            content_start = match.span()[1]
            content_end = len(text)
            break

    if metadata_kind == MetadataKind.Yaml:
        metadata = load_yaml(text[meta_start:meta_end])
    elif metadata_kind == MetadataKind.Toml:
        metadata = toml.loads(text[meta_start:meta_end])
    else:
        # We didn't find any metadata here, so everything must be content
        return {}, text, 1

    content = text[content_start:content_end]
    # +2 for the start/end marker, which is excluded from the
    # meta_start/meta_end
    # +1 because we start counting at 1
    return metadata, content, text[meta_start:meta_end].count('\n') + 3


def fixup_relative_links(document: 'DocumentNode'):
    """Replace relative links in the document with links relative to the
    site root."""
    if not document.content:
        return

    # early out if there's no relative link in here, as the parsing is
    # very expensive
    if "href=\"." not in document.content:
        return

    from bs4 import BeautifulSoup
    soup = BeautifulSoup(document.content, 'lxml')

    def is_relative_url(s):
        return s and s[0] == '.'

    for link in soup.find_all('a', {'href': is_relative_url}):
        target = link.attrs.get('href')
        link.attrs['href'] = \
            str(document.path.parent / pathlib.PurePosixPath(target))

    document.content = str(soup)


def fixup_date(document: 'DocumentNode'):
    """If the date in the document is a string, try to parse it to produce a
    datetime object."""
    if 'date' in document.metadata:
        date = document.metadata['date']
        if isinstance(date, str):
            document.metadata['date'] = dateparser.parse(date)


class FixupDateTimezone:
    """Set the timezone of the ``metadata['date']`` field to the local timezone
    if no timezone has been set."""
    def __init__(self):
        import tzlocal
        self.__tz = tzlocal.get_localzone()

    def __call__(self, document: 'DocumentNode'):
        '''If the date in the document has no timezone info, set it to the
        local timezone.'''
        if 'date' in document.metadata:
            date = document.metadata['date']
            if date.tzinfo is None:
                document.metadata['date'] = date.replace(tzinfo=self.__tz)


class DocumentNode(Node):
    _load_fixups: List[Callable]
    """These functions are called right after the document has been loaded,
    and can be used to fixup metadata, content, etc. before it gets processed
    (These should be called before :py:meth:`load`/:py:meth:`reload`
    returns.)"""

    _process_fixups: List[Callable]
    """These functions are called after a document has been processed
    (These should be called before :py:meth:`process` returns)."""

    content: str | None

    def __init__(self, src, path: pathlib.PurePosixPath,
                 metadata_path: Optional[pathlib.Path] = None):
        super().__init__()
        self.kind = NodeKind.Document
        self.src = src
        self.path = path
        self.metadata_path = metadata_path
        self.content = None
        self._load_fixups = []
        self._process_fixups = []
        self._content_line_start = None

    def set_fixups(self, *, load_fixups, process_fixups) -> None:
        """Set the fixups that should be applied to this document node.
        The fixups should be set *before* calling :py:meth:`load`.

        :param load_fixups: These functions will be executed before
                            :py:meth:`load` returns.
        :param process_fixups: These functions will be executed before
                            :py:meth:`process` returns.
        """
        self._load_fixups = load_fixups
        self._process_fixups = process_fixups

    def load(self):
        """Load the content of this node."""
        self._load()
        self._apply_load_fixups()
        document_loaded.send(self, document=self, content=self._raw_content)

    def validate_metadata(self):
        if self.metadata is None:
            raise Exception(f"No metadata for document: '{self.src}'")
        if 'title' not in self.metadata:
            raise Exception(f"'title' missing for Document: '{self.src}'")

    def _apply_load_fixups(self):
        for fixup in self._load_fixups:
            fixup(self)

    def _apply_process_fixups(self):
        for fixup in self._process_fixups:
            fixup(self)

    def _load(self):
        assert self.src

        if self.metadata_path:
            self.metadata = load_yaml(self.metadata_path.read_text())
            self._raw_content = self.src.read_text('utf-8')
            self._content_line_start = 1
        else:
            self.metadata, self._raw_content, self._content_line_start = \
                extract_metadata_content(self.src.read_text('utf-8'))

    def reload(self):
        """Reload this node from disk.

        By default, this just forwards to :py:meth:`_load`.
        """
        self._load()
        self._apply_load_fixups()

    def publish(self, publisher: Publisher) -> pathlib.Path:
        """Publish this node using the provided publisher."""
        return publisher.publish_document(self)


class HtmlDocumentNode(DocumentNode):
    """A node representing a Html document."""

    def process(self, cache: Cache, **kwargs):
        self.content = self._raw_content

        self._apply_process_fixups()


class MarkdownDocumentNode(DocumentNode):
    """A node representing a Markdown document."""
    def __init__(self, configuration, **kwargs):
        super().__init__(**kwargs)
        self.__md = self._create_markdown_processor(configuration)

    def _create_markdown_processor(self, configuration):
        from markdown import Markdown
        from .md import LiaraMarkdownExtensions

        self.__mdext = LiaraMarkdownExtensions(self)

        extensions = [
            self.__mdext
        ] + configuration['content.markdown.extensions']

        extension_configs = configuration['content.markdown.config']
        output = configuration['content.markdown.output']

        return Markdown(extensions=extensions,
                        extension_configs=extension_configs,
                        output_format=output)

    def process(self, cache: Cache, **kwargs):
        import hashlib
        from .md import ShortcodeException

        if '$data' in kwargs:
            self.__mdext.set_data(kwargs['$data'])

        byte_content = self._raw_content.encode('utf-8')
        content_hash = hashlib.sha256(byte_content).digest()
        if content := cache.get(content_hash):
            assert isinstance(content, str)
            self.content = content
            return

        # We re-raise shortcode exceptions to adjust the line number to
        # make debugging easier
        try:
            self.content = self.__md.convert(self._raw_content)
        except ShortcodeException as sx:
            raise sx.with_line_offset(self._content_line_start) from sx
        finally:
            self.__md.reset()

        self._apply_process_fixups()

        cache.put(content_hash, self.content)


class DataNode(Node):
    """A data node.

    Data nodes consist of a dictionary. This can be used to store arbitrary
    data as part of a :py:class:`liara.site.Site`, and make it available to
    templates (for instance, a menu structure could go into a data node.)
    """
    def __init__(self, src: pathlib.Path, path: pathlib.PurePosixPath):
        super().__init__()
        self.kind = NodeKind.Data
        self.src = src
        self.path = path
        self.content = load_yaml(self.src.open('rb'))


class IndexNode(Node):
    """An index node.

    Index nodes are created for every folder if there is no ``_index`` node
    present, and from indices. An index node can optionally contain a list of
    references, in case the referenced nodes by this index are not direct
    children of this node.
    """

    references: List[Node]
    """Nodes referenced by this index node.

    An index can not rely on using ``children`` as those have to be below the
    path of the parent node. The ``references`` list allows to reference nodes
    elsewhere in the site."""

    def __init__(self, path: pathlib.PurePosixPath,
                 metadata: Optional[Dict] = None):
        super().__init__()
        self.kind = NodeKind.Index
        self.src = None
        self.path = path
        self.metadata = metadata if metadata else {}
        self.references = []

    def add_reference(self, node):
        """Add a reference to an arbitrary node in the site."""
        self.references.append(node)

    def publish(self, publisher) -> pathlib.Path:
        """Publish this node using the provided publisher."""
        return publisher.publish_index(self)


class GeneratedNode(Node):
    def __init__(self, path: pathlib.PurePosixPath,
                 metadata: Optional[Dict] = None):
        super().__init__()
        self.kind = NodeKind.Generated
        self.src = None
        self.path = path
        self.metadata = metadata if metadata else {}
        self.content: Optional[Union[bytes, str]] = None

    def generate(self) -> None:
        """Generate the content of this node.

        After this function has finished, ``self.content`` must be populated
        with the generated content."""
        pass

    def publish(self, publisher: Publisher):
        """Publish this node using the provided publisher."""
        return publisher.publish_generated(self)


_REDIRECTION_TEMPLATE = """<!DOCTYPE HTML>
<html lang="en-US">
    <head>
        <meta charset="UTF-8">
        <meta http-equiv="refresh" content="0; url={{NEW_URL}}">
        <script type="text/javascript">
            window.location.href = "{{NEW_URL}}"
        </script>
        <title>Page Redirection</title>
    </head>
    <body>
        <h1>Page has been moved</h1>
        <p>If you are not redirected automatically, follow this
        <a href='{{NEW_URL}}'>link.</a>.</p>
    </body>
</html>"""


class RedirectionNode(GeneratedNode):
    """A redirection node triggers a redirection to another page.

    This node gets processed into a simple web site which tries to redirect
    using both ``<meta http-equiv="refresh">`` and Javascript code setting
    ``window.location``.
    """

    def __init__(self,
                 path: pathlib.PurePosixPath,
                 dst: pathlib.PurePosixPath,
                 *,
                 base_url=''):
        super().__init__(path)
        self.dst = dst
        self.__base_url = base_url

    def generate(self):
        text = _REDIRECTION_TEMPLATE.replace('{{NEW_URL}}',
                                             self.__base_url
                                             + self.dst.as_posix())
        self.content = text


class ResourceNode(Node):
    """A resource node applies some process when creating the output.

    This is useful if you have content where the source cannot be interpreted,
    and requires some process first before it becomes usable -- for instance,
    ``SASS`` to ``CSS`` compilation.
    """
    def __init__(self, src, path: pathlib.PurePosixPath, metadata_path=None):
        super().__init__()
        self.kind = NodeKind.Resource
        self.src = src
        self.path = path
        self.content = None
        if metadata_path:
            self.metadata = load_yaml(open(metadata_path, 'r'))

    def reload(self) -> None:
        pass

    def publish(self, publisher: Publisher) -> pathlib.Path:
        """Publish this node using the provided publisher."""
        return publisher.publish_resource(self)


_SASS_COMPILER = Literal['cli', 'libsass']


class _AsyncSassTask(_AsyncTask):
    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, cache_key: bytes,
                 compiler: _SASS_COMPILER,
                 src: pathlib.Path):
        self.__cache_key = cache_key
        self.__src = src
        self.__compiler = compiler

    def _compile_using_cli(self):
        import subprocess
        import sys

        self.__log.debug('Processing "%s" using "sass" binary', self.__src)

        result = subprocess.run(
            ['sass', str(self.__src)],
            # On Windows, we need to set shell=True, otherwise, the
            # sass binary installed using npm install -g sass won't
            # be found.
            shell=sys.platform == 'win32',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)

        try:
            result.check_returncode()
        except FileNotFoundError:
            self.__log.error('Could not find "sass" binary. Make sure that '
                             'SASS is installed and available in your PATH.')
            raise
        except Exception:
            self.__log.error('SASS compilation of file "%s" failed: "%s"',
                             self.__src,
                             result.stderr.decode('utf-8'))
            raise

        self.__log.debug('Done processing "%s"', self.__src)
        return result.stdout

    def _compile_using_libsass(self):
        import sass
        self.__log.debug(f'Processing "{self.__src}" using "libsass"')

        result = sass.compile(filename=str(self.__src)).encode('utf-8')

        self.__log.debug(f'Done processing "{self.__src}"')

        return result

    def process(self):
        try:
            if self.__compiler == 'cli':
                return self._compile_using_cli()
            elif self.__compiler == 'libsass':
                return self._compile_using_libsass()
        except Exception as e:
            self.__log.warning(f'Failed to compile SCSS file "{self.__src}"',
                               exc_info=e)

    def update_cache(self, content: object, cache: Cache):
        cache.put(self.__cache_key, content)


class SassResourceNode(ResourceNode):
    """This resource node compiles ``.sass`` and ``.scss`` files to CSS
    when built.
    """
    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, src, path: pathlib.PurePosixPath, metadata_path=None):
        super().__init__(src, path, metadata_path)
        if src.suffix not in {'.scss', '.sass'}:
            raise Exception("SassResource can be only created for a .scss or "
                            " .sass file")

        self.path = self.path.with_suffix('.css')
        self.__compiler: _SASS_COMPILER = 'cli'

    def set_compiler(self, compiler: _SASS_COMPILER):
        self.__compiler = compiler

    def reload(self) -> None:
        self.content = None

    def process(self, cache: Cache, **kwargs):
        import hashlib
        if self.content is not None:
            return

        assert self.src
        cache_key = hashlib.sha256(self.src.open('rb').read()).digest()

        if (value := cache.get(cache_key)) is not None:
            self.content = value
            return

        return _AsyncSassTask(cache_key, self.__compiler, self.src)


NT = TypeVar('NT', bound=Node)


class NodeFactory(Generic[NT]):
    """A generic factory for nodes, which builds nodes based on the file
    type."""
    __known_types: Dict[str, Tuple[Type, Any]]

    def __init__(self):
        self.__known_types = {}

    @property
    def known_types(self):
        return self.__known_types.keys()

    def register_type(self,
                      suffixes: Union[str, Iterable[str]],
                      node_type: type,
                      *,
                      extra_args: Dict[str, Any] = dict()) -> None:
        """Register a new node type.

        :param suffixes: Either one suffix, or a list of suffixes to be
                         registered for this type. For instance, a node
                         representing an image could be registered to
                         ``[.jpg, .png, .webp]``.
        :param node_type: The type of the node to be created.
        :param extra_args: A dictionary of additional arguments for the
                           node constructor. This dictionary gets passed
                           as ``**kwargs`` into the node constructor.
        """
        if isinstance(suffixes, str):
            suffixes = [suffixes]

        for suffix in suffixes:
            self.__known_types[suffix] = (node_type, extra_args,)

    def _create_node(self,
                     cls,
                     src: pathlib.Path,
                     path: pathlib.PurePosixPath,
                     metadata_path: Optional[pathlib.Path] = None,
                     **kwargs) -> NT:
        """This is the actual creation function.

        :param cls: The class of the node to instantiate.
        :param src: The source file path.
        :param path: The output path.
        :param metadata_path: The path to a metadata file.
        :param kwargs: Any additional arguments are passed though.
        :return: An instance of ``cls``.

        Derived classes can use this function to customize the node creation.
        This call uses named arguments so derived classes can easily extract
        arguments specific to them.
        """
        # We call the constructor with all keyword arguments which makes it
        # easier to manipulate the argument list/remove arguments
        return cls(src=src, path=path, metadata_path=metadata_path, **kwargs)

    def create_node(self, suffix: str,
                    src: pathlib.Path,
                    path: pathlib.PurePosixPath,
                    metadata_path: Optional[pathlib.Path] = None) -> NT:
        """Create a node using the provided parameters."""
        cls, extra_args = self.__known_types[suffix]
        node = self._create_node(cls, src, path, metadata_path, **extra_args)
        return self._on_node_created(node)

    def _on_node_created(self, node: NT):
        """Called after a node has been created.

        This can be used to further configure the node from the factory
        before returning it from the factory, for instance, to pass
        configuration down into the node.

        .. versionadded:: 2.3.4"""
        return node


class ResourceNodeFactory(NodeFactory[ResourceNode]):
    """A factory for resource nodes."""

    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, configuration):
        super().__init__()
        self.register_type(['.sass', '.scss'], SassResourceNode)

        self.__sass_compiler = configuration['build.resource.sass.compiler']
        if self.__sass_compiler == 'libsass':
            self.__log.info(
                'Support for "libsass" as the compiler for SASS '
                'files is deprecated and will be removed in a future release. '
                'Please check the documentation how to use the SASS '
                'command line compiler.')

    def _on_node_created(self, node: NT):
        if isinstance(node, SassResourceNode):
            node.set_compiler(self.__sass_compiler)
        return node


class DocumentNodeFactory(NodeFactory[DocumentNode]):
    """A factory for document nodes."""
    def __setup_fixups(self, configuration):
        if configuration['relaxed_date_parsing']:
            # This is tricky, as fixup_date depends on this running
            # first. We thus prepend this before any other fixup and hope this
            # is the only one with ordering issues.
            self.__load_fixups.insert(0, fixup_date)
        if configuration['allow_relative_links']:
            self.__process_fixups.append(fixup_relative_links)

    def __init__(self, configuration):
        super().__init__()
        self.__load_fixups: List[Callable] = [FixupDateTimezone()]
        self.__process_fixups = []

        self.__setup_fixups(configuration)

        self.register_type(
            ['.md'], MarkdownDocumentNode,
            extra_args={
                'configuration': configuration
            })
        self.register_type(['.html'], HtmlDocumentNode)

    def _on_node_created(self, node):
        node.set_fixups(
            load_fixups=self.__load_fixups,
            process_fixups=self.__process_fixups)
        node.load()
        return node


class StaticNode(Node):
    """A static data node.

    Static nodes are suitable for large static data which never changes, for
    instance, binary files, videos, images etc.
    """
    src: pathlib.Path   # Unlike a generic Node, a StaticNode always has a
                        # source

    def __init__(self, src: pathlib.Path, path: pathlib.PurePosixPath,
                 metadata_path=None):
        super().__init__()
        self.kind = NodeKind.Static
        self.src = src
        self.path = path
        if metadata_path:
            self.metadata = load_yaml(open(metadata_path, 'r'))

    def update_metadata(self) -> None:
        """Update metadata by deriving some metadata from the source file,
        if possible.

        For static nodes pointing to images, this will create a new metadata
        field ``image_size`` and populate it with the image resolution."""
        from PIL import Image
        if self.is_image:
            image = Image.open(self.src)
            self.metadata.update({
                'image_size': image.size
            })

    @property
    def is_image(self):
        """Return ``True`` if this static file is pointing to an image."""
        return self.src.suffix in {'.jpg', '.png', '.webp'}

    def publish(self, publisher: Publisher) -> pathlib.Path:
        """Publish this node using the provided publisher."""
        return publisher.publish_static(self)


class _AsyncThumbnailTask(_AsyncTask):
    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, cache_key: bytes, src: pathlib.Path,
                 size: Dict[str, int],
                 format: Optional[str]):
        self.__src = src
        self.__size = size
        self.__cache_key = cache_key
        assert format != 'original'
        self.__format = format

    def process(self):
        from PIL import Image
        import io
        self.__log.debug('Processing "%s"', self.__src)
        image = Image.open(self.__src)
        width, height = image.size

        scale = 1
        if 'height' in self.__size:
            scale = min(self.__size['height'] / height, scale)
        if 'width' in self.__size:
            scale = min(self.__size['width'] / width, scale)
        width *= scale
        height *= scale

        image.thumbnail((int(width), int(height),))
        storage = io.BytesIO()
        assert self.__src
        result = None

        format_from_suffix = {
            '.jpg': 'JPEG',
            '.png': 'PNG',
            '.webp': 'WEBP'
        }

        if self.__format:
            format = self.__format.upper()
        else:
            format = format_from_suffix.get(self.__src.suffix.lower())

        if format is None:
            raise Exception("Unsupported image type for thumbnails")

        image.save(storage, format)
        result = storage.getvalue()

        self.__log.debug('Done processing "%s"', self.__src)
        return result

    def update_cache(self, content: object, cache: Cache):
        cache.put(self.__cache_key, bytes(content))


class ThumbnailNode(ResourceNode):
    def __init__(self, src: pathlib.Path,
                 path: pathlib.PurePosixPath, size: Dict[str, int],
                 format: str | None = 'original'):
        super().__init__(src, path)
        self.__size = size
        self.__format = format

    def __get_cache_key(self) -> bytes:
        import hashlib
        assert self.src
        cache_key = hashlib.sha256(self.src.open('rb').read()).digest()

        if 'height' in self.__size:
            cache_key += self.__size['height'].to_bytes(4, 'little')
        else:
            cache_key += bytes([0, 0, 0, 0])

        if 'width' in self.__size:
            cache_key += self.__size['width'].to_bytes(4, 'little')
        else:
            cache_key += bytes([0, 0, 0, 0])

        if self.__format:
            cache_key += self.__format.encode('utf-8')

        return cache_key

    def process(self, cache: Cache, **kwargs) -> _AsyncThumbnailTask | None:
        cache_key = self.__get_cache_key()
        if content := cache.get(cache_key):
            self.content = content
            return None

        assert self.src

        async_task = _AsyncThumbnailTask(cache_key, self.src,
                                         self.__size,
                                         self.__format)

        return async_task


def _parse_node_kind(kind) -> NodeKind:
    """Parse a node kind from a string.

    This allows certain shortcuts to be used in addition to the exact node kind
    name.

    .. versionadded:: 2.4"""
    kind_map = {
            'document': NodeKind.Document,
            'index': NodeKind.Index,
            'static': NodeKind.Static,
            'generated': NodeKind.Generated,
            'resource': NodeKind.Resource,
            'data': NodeKind.Data,

            # Additional shortcuts
            'doc': NodeKind.Document,
            'idx': NodeKind.Index,
    }

    return kind_map[kind]


def _process_node_sync(node: Union[ResourceNode, DocumentNode], cache: Cache, **kwargs):
    """
    Process a node synchronously, even if it returned an async result.
    """
    if task := node.process(cache, **kwargs):
        node.content = task.process()
        task.update_cache(node.content, cache)



from typing import (
    IO,
    Optional,
    Text,
    Union
)


def load_yaml(s: Union[bytes, IO, IO[bytes], Text, IO[Text]]):
    """Load a Yaml document.

    This is a helper function which tries to use the fast ``CLoader``
    implementation and falls back to the native Python version on failure.
    """
    import yaml
    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader
    return yaml.load(s, Loader=Loader)


def dump_yaml(data, stream: Optional[IO] = None):
    """Dump an object to Yaml.

    This is a helper function which tries to use the fast ``CDumper``
    implementation and falls back to the native Python version on failure.
    """
    import yaml
    try:
        from yaml import CDumper as Dumper
    except ImportError:
        from yaml import Dumper

    return yaml.dump(data, stream, Dumper=Dumper)



from typing import (
    Dict,
    Any,
)


def create_default_configuration() -> Dict[str, Any]:
    """Creates a dictionary with the default configuration."""
    # When changing this, make sure to update the docs as well
    return {
        'content_directory': 'content',
        'resource_directory': 'resources',
        'static_directory': 'static',
        'output_directory': 'output',
        'generator_directory': 'generators',
        'plugin_directories': [],
        'build': {
            'clean_output': True,
            'cache.fs.directory': 'cache',
            'cache.db.directory': 'cache',
            'cache.redis.host': 'localhost',
            'cache.redis.port': 6379,
            'cache.redis.db': 0,
            'cache.redis.expiration_time': 60,
            'cache.type': 'fs',
            'resource.sass.compiler': 'libsass'
        },
        'ignore_files': ['*~'],
        'content': {
            'filters': ['date', 'status'],
            'markdown': {
                'extensions': [
                    'pymdownx.arithmatex',
                    'fenced_code',
                    'codehilite',
                    'smarty',
                    'tables',
                    'admonition'
                ],
                'config': {
                    'codehilite': {
                        'css_class': 'code'
                    },
                    'pymdownx.arithmatex': {
                        'generic': True
                    }
                },
                'output': 'html5'
            }
        },
        'template': 'templates/default.yaml',
        'routes': {
            'static': 'static_routes.yaml'
        },
        'collections': 'collections.yaml',
        'relaxed_date_parsing': False,
        'allow_relative_links': True,
        'feeds': 'feeds.yaml',
        'metadata': 'metadata.yaml'
    }


def create_default_template_configuration() -> Dict[str, Any]:
    """Creates the default template configuration."""
    return {
        'backend': 'jinja2',
        'backend_options': {
            'jinja2': {
                'trim_blocks': True,
                'lstrip_blocks': True
            }
        }
    }


def create_default_metadata() -> Dict[str, Any]:
    """Creates a dictionary with the default metadata."""
    return {
        'title': 'Default title',
        'base_url': 'https://example.org',
        'description': 'Default description',
        'language': 'en-US',
        'copyright': 'Licensed under the '
                     '<a href="http://creativecommons.org/licenses/by/4.0/"> '
                     'Creative Commons Attribution 4.0 International License.'
    }



from .cache import Cache
import pathlib
from typing import (
    Any,
    Dict,
    Optional,
)
from types import MappingProxyType

from abc import abstractmethod, ABC

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .query import Query
    from . import Site


def _match_url(url: str | pathlib.PurePosixPath, pattern: str, site: 'Site') \
        -> Optional[int]:
    """Match an url against a pattern.

    :return: An integer indicating the match score, with 0 being a perfect
             match and higher values being increasingly bad. ``None`` is
             returned if no match was found.
    """
    import fnmatch
    import urllib.parse
    from .nodes import _parse_node_kind
    if '?' in pattern and site:
        pattern, params_str = pattern.split('?')

        node = site.get_node(url)
        assert node
        params = urllib.parse.parse_qs(params_str)

        if kinds := params.get('kind'):
            kinds = {_parse_node_kind(kind) for kind in kinds}

            if node.kind not in kinds:
                return None

    # Exact matches always win
    if pattern == str(url):
        return 0
    # If not exact, we'll look for the longest matching pattern,
    # assuming it is the most specific
    if fnmatch.fnmatch(str(url), pattern):
        # abs is required, if our pattern is /*, and the url we match against
        # is /, then the pattern is longer than the URL
        return abs(len(str(url)) - len(pattern))

    return None


class Template(ABC):
    def __init__(self, template_path='<unknown>'):
        self._template_path = template_path

    @property
    def path(self):
        return self._template_path

    @abstractmethod
    def render(self, **kwargs) -> str:
        ...


class TemplateRepository(ABC):
    def __init__(self, paths: Dict[str, str]):
        self.__paths = paths

    def update_paths(self, paths: Dict[str, str]):
        self.__paths = paths

    @abstractmethod
    def find_template(self, url: pathlib.PurePosixPath, site: 'Site') \
            -> Template:
        ...

    def _match_template(self, url: pathlib.PurePosixPath, site: 'Site') -> tuple[str, str]:
        """Match a path to a template, if any, otherwise throw an exception.

        Returns a tuple containing the matching template name, and the pattern
        that matched.
        """
        best_match = None
        best_pattern = None
        best_score = None
        longest_matching_pattern_length = -1
        for pattern, template in self.__paths.items():
            score = _match_url(url, pattern, site)
            if score is None:
                continue

            # If the pattern is a better match, we always update
            if best_score is None or score < best_score:
                best_score = score
                best_match = template
                best_pattern = pattern
                longest_matching_pattern_length = len(pattern)

            # Tie breaker: The longer pattern wins
            if best_score == score:
                if len(pattern) > longest_matching_pattern_length:
                    best_match = template
                    best_pattern = pattern
                    longest_matching_pattern_length = len(pattern)

        if not best_match:
            raise Exception(f'Could not find matching template for path: '
                            f'"{url}"')

        return best_match, best_pattern


class MakoTemplate(Template):
    def __init__(self, template, template_path):
        super().__init__(template_path)
        self.__template = template

    def render(self, **kwargs) -> str:
        return self.__template.render(**kwargs)


class MakoTemplateRepository(TemplateRepository):
    def __init__(self, paths: Dict[str, str], path: pathlib.Path):
        super().__init__(paths)
        from mako.lookup import TemplateLookup
        self.__lookup = TemplateLookup(directories=[str(path)])

    def find_template(self, url: pathlib.PurePosixPath, site: 'Site') \
            -> Template:
        template, _ = self._match_template(url, site)
        return MakoTemplate(self.__lookup.get_template(template), template)


class Jinja2Template(Template):
    def __init__(self, template, template_path):
        super().__init__(template_path)
        self.__template = template

    def render(self, **kwargs) -> str:
        return self.__template.render(**kwargs)


class Jinja2TemplateRepository(TemplateRepository):
    """Jinja2 based template repository."""
    def __init__(self, paths: Dict[str, str], path: pathlib.Path,
                 cache: Optional[Cache] = None, *,
                 options: Optional[Dict[str, Any]] = None):
        super().__init__(paths)
        self.__path = path
        self.__cache = cache
        self.__create_environment(options)

    def __create_environment(self, options: Optional[Dict[str, Any]] = None):
        from jinja2 import FileSystemLoader, Environment, BytecodeCache
        from .util import readtime
        import io

        def sanitize_options(options):
            if not options:
                return dict()

            forbidden_options = {
                'undefined',
                'finalize',
                'loader',
                'bytecode_cache',
                'enable_async'
            }

            return {k: v for k, v in options.items()
                    if k not in forbidden_options}

        class Jinja2BytecodeCache(BytecodeCache):
            def __init__(self, cache: Cache):
                self.__cache = cache

            def clear(self):
                pass

            def load_bytecode(self, bucket):
                key = bucket.key.encode('utf-8')
                if content := self.__cache.get(key):
                    assert isinstance(content, bytes)
                    s = io.BytesIO(content)
                    bucket.load_bytecode(s)

            def dump_bytecode(self, bucket):
                key = bucket.key.encode('utf-8')
                s = io.BytesIO()
                bucket.write_bytecode(s)

                # Caches pickle under the hood, and a memoryview cannot be
                # pickled -- we thus convert to bytes here
                self.__cache.put(key, s.getbuffer().tobytes())

        if self.__cache:
            cache = Jinja2BytecodeCache(self.__cache)
        else:
            cache = None

        self.__env = Environment(
            loader=FileSystemLoader(str(self.__path)),
            bytecode_cache=cache,
            **sanitize_options(options))
        self.__env.filters['readtime'] = readtime

    def find_template(self, url: pathlib.PurePosixPath, site: 'Site') \
            -> Template:
        template, _ = self._match_template(url, site)
        return Jinja2Template(self.__env.get_template(template), template)


class Page:
    """A wrapper around :py:class:`~liara.nodes.DocumentNode` and
    :py:class:`~liara.nodes.IndexNode` for use inside templates.

    Templates only get applied to those node types, and the :py:class:`Page`
    class provides convenience accessors while hiding the underlying node from
    template code.
    """
    def __init__(self, node):
        self.__node = node

    @property
    def content(self) -> str:
        """Provides the content of this page.
        """
        return getattr(self.__node, 'content', '')

    @property
    def url(self) -> str:
        """Provides the current path of this page.
        """
        # __node.path is a PosixPath object, but inside a template we want to
        # use a basic string
        return str(self.__node.path)

    @property
    def meta(self) -> MappingProxyType[str, Any]:
        """Provides the metadata associated with this page.

        .. deprecated:: 2.1.2
           Use :py:attr:`~liara.template.Page.metadata` instead.
        .. versionchanged:: 2.6.2 Made read-only
        """
        return self.__node.metadata

    @property
    def metadata(self) -> MappingProxyType[str, Any]:
        """Provides the metadata associated with this page.

        .. versionadded:: 2.1.2
        .. versionchanged:: 2.6.2 Made read-only
        """
        return self.__node.metadata

    @property
    def _node(self):
        return self.__node

    def __str__(self):
        return f'Page({self.url})'

    @property
    def references(self) -> 'Query':
        """Provides the list of referenced nodes by this page.

        This can be only used if the current page is an
        :py:class:`~liara.nodes.IndexNode`, in all other cases this will fail.
        For index nodes, this will return the list of references as a
        :py:class:`~liara.query.Query` instance.
        """

        from .nodes import NodeKind
        from .query import Query
        assert self.__node.kind == NodeKind.Index

        return Query(self.__node.references)

    @property
    def children(self) -> 'Query':
        """Return all child pages of this page, i.e. document and index
        nodes.

        .. versionadded:: 2.4"""
        from .query import Query
        return Query(self.__node.children).with_node_kinds('doc', 'idx')


class SiteTemplateProxy:
    """A wrapper around :py:class:`Site` for use inside templates.
    """
    __site: 'Site'

    def __init__(self, site: 'Site'):
        self.__site = site
        self.__data = site.merged_data

    @property
    def data(self) -> MappingProxyType[str, Any]:
        """Get the union of all :py:class:`liara.nodes.DataNode`
        instances in this site.

        .. versionchanged:: 2.6.2 Made read-only
           Before 2.6.2, this dictionary could be modified, but that was always
           an unsafe operation.
        """
        return self.__data

    @property
    def metadata(self) -> MappingProxyType[str, Any]:
        """Provide access to the metadata of this site.

        .. versionchanged:: 2.6.2 Made read-only
           Before 2.6.2, this dictionary could be modified, but that was always
           an unsafe operation.
        """
        return MappingProxyType(self.__site.metadata)

    def select(self, query) -> 'Query':
        """Run a query on this site. Returns any node matching the query.
        """
        from .query import Query
        return Query(self.__site.select(query))

    def select_pages(self, query) -> 'Query':
        """Run a query on this site and return only matching pages, i.e.
        document and index nodes.

        .. versionadded:: 2.4"""
        return self.select(query).with_node_kinds('doc', 'idx')

    def get_page_by_url(self, url) -> Optional[Page]:
        """Return a page by URL. If the page cannot be found, return
        ``None``."""
        node = self.__site.get_node(pathlib.PurePosixPath(url))

        if node:
            return Page(node)
        else:
            return None

    def get_previous_in_collection(self, collection: str, page: Page) \
            -> Optional[Page]:
        """Given a collection and a page, return the previous page in this
        collection or ``None`` if this is the first page.
        """
        previous_node = self.__site.get_previous_in_collection(collection,
                                                               page._node)
        if previous_node:
            return Page(previous_node)
        else:
            return None

    def get_next_in_collection(self, collection: str, page: Page) \
            -> Optional[Page]:
        """Given a collection and a page, return the next page in this
        collection or ``None`` if this is the last page.
        """
        next_node = self.__site.get_next_in_collection(collection, page._node)
        if next_node:
            return Page(next_node)
        else:
            return None

    def get_collection(self, collection: str) -> 'Query':
        """Get a collection in form of a :py:class:`liara.query.Query` for
        further filtering/sorting.
        """
        from .query import Query
        # We need to turn this into a list as nodes is of type dict_values
        # and can't be sorted/reversed otherwise
        return Query(list(self.__site.get_collection(collection).nodes))



from liara import signals


def register():
    signals.commandline_prepared.connect(_register_cli)


def _register_cli(sender, cli):
    from liara.cmdline import pass_environment
    import sys

    @cli.command()
    @pass_environment
    def has_pending_document(env):
        """Check if there is a pending document.

        If a pending document is found, the return code from the application
        will be 1."""
        documents_filtered_by_date = 0

        def on_content_filtered(sender, **kw):
            nonlocal documents_filtered_by_date
            if kw['filter'].name == 'date':
                documents_filtered_by_date += 1

        signals.content_filtered.connect(on_content_filtered)

        env.liara.discover_content()

        if documents_filtered_by_date > 0:
            print(f'{documents_filtered_by_date} document(s) pending')
            sys.exit(1)



from typing import Iterable, List, Iterator, Optional
from .nodes import Node, NodeKind, _parse_node_kind
from .template import Page
from typing import Union
from abc import abstractmethod, ABC
import logging

import re


class SelectionFilter(ABC):
    """Base class for query selection filters."""
    @abstractmethod
    def match(self, node: Node) -> bool:
        """Return ``True`` if the node should be kept, else ``False``."""
        ...


class MetadataFilter(SelectionFilter):
    """Filter items which contain a specific metadata field and optionally
    check if that field matches the provided value.
    """
    def __init__(self, name, value=None):
        self.__name = name
        self.__value = value

    def match(self, node: Node) -> bool:
        if self.__name in node.metadata:
            if self.__value is not None:
                return node.metadata[self.__name] == self.__value
            else:
                return True
        return False


class TagFilter(SelectionFilter):
    """Filter items by a specific tag, this expects a metadata field named
    ``tags`` to be present, and that field must support checks for containment
    using ``in``."""
    def __init__(self, name):
        self.__name = name

    def match(self, node: Node) -> bool:
        tags = node.metadata.get('tags')
        if tags is not None:
            return self.__name in tags
        return False


class ExcludeFilter(SelectionFilter):
    """Filter items by a provided pattern. The pattern is matched against the
    path. If it matches, the item will be ignored."""
    def __init__(self, pattern):
        self.__pattern = re.compile(pattern)

    def match(self, node: Node) -> bool:
        path = str(node.path)
        return self.__pattern.search(path) is None


class NodeKindFilter(SelectionFilter):
    """Filter items based on the node kind. Use exclude to invert.

    .. versionadded:: 2.4"""
    def __init__(self, kinds, *, exclude=False):
        self.__kinds = {_parse_node_kind(kind) for kind in kinds}
        self.__exclude = exclude

    def match(self, node: Node) -> bool:
        if self.__exclude:
            return node.kind not in self.__kinds
        return node.kind in self.__kinds


class Sorter(ABC):
    """Base class for query sorters."""
    def __init__(self, reverse=False):
        self._reverse = reverse

    @abstractmethod
    def get_key(self, item: Node):
        """Return the key to be used for sorting."""
        ...

    @property
    def reverse(self) -> bool:
        """Returns ``True`` if the sort order should be reversed."""
        return self._reverse


class MetadataSorter(Sorter):
    """Sort nodes by metadata."""
    def __init__(self, item: str, reverse=False, case_sensitive=False):
        super().__init__(reverse)
        self.__item = item
        self.__case_sensitive = case_sensitive

    def get_key(self, item: Node):
        key = item.metadata.get(self.__item)
        # We check if the key is None here, as we can't sort None with anything
        # nor does it make much sense (i.e., if you sort by title, and an
        # item has no title, that's probably because it was incorrectly
        # included, not because someone wanted items without a title
        # first/last)
        if key is None:
            raise Exception(f'Cannot sort node "{item.path}" by key '
                            f'"{self.__item}" as the node is missing that '
                            'key from its metadata.')
        if isinstance(key, str) and self.__case_sensitive is False:
            return key.lower()
        return key


class Query(Iterable[Union[Node, Page]]):
    """A query modifies a list of nodes, by sorting and filtering entries.

    Sorting requires the sort key to be present on all nodes that are to be
    sorted, otherwise an error is raised. Nodes without a particular key can
    be filtered out using :py:meth:`~liara.query.Query.with_metadata`.

    Index and document nodes will be wrapped in a
    :py:class:`~liara.template.Page` instance. Everything else will be returned
    as a :py:class:`~liara.nodes.Node`.
    """

    __filters: List[SelectionFilter]
    __nodes: List[Node]
    __sorters: List[Sorter]
    __limit: int
    __reversed: bool
    __result: Optional[List[Union[Node, Page]]]

    __log = logging.getLogger(f'{__name__}.{__qualname__}')

    def __init__(self, nodes: Iterable[Node]):
        """Create a query object for the list of specified nodes."""
        self.__nodes = list(nodes)
        self.__limit = -1
        self.__filters = []
        self.__sorters = []
        self.__reversed = False
        self.__result = None

    def limit(self, limit: int) -> 'Query':
        """Limit this query to return at most ``limit`` results."""
        self.__limit = limit
        return self

    def with_metadata(self, name, value=None) -> 'Query':
        """Limit this query to only include nodes which contain the specific
        metadata field.

        :param value: If ``value`` is provided, the field must exist and match
                      the provided value.
        """
        self.__filters.append(MetadataFilter(name, value))
        return self

    def with_tag(self, name) -> 'Query':
        """Limit this query to only include nodes with a metadata field named
        ``tags`` which contains the specified tag name.
        """
        self.__filters.append(TagFilter(name))
        return self

    def exclude(self, pattern) -> 'Query':
        """Exclude nodes matching the provided regex pattern. The pattern will
        be applied to the full path.
        
        .. note::
        
          This means that a pattern of ``a`` for example will match a path like
          ``bar``, as ``a`` is found inside ``bar``. If you want to match the
          specific regular expression and not allow for matches within,
          use ``^/a$``.
        """
        self.__filters.append(ExcludeFilter(pattern))
        return self

    def without_node_kinds(self, *args) -> 'Query':
        """Excludes nodes of a specific type. Multiple kinds can be passed in.

        .. versionadded:: 2.4"""
        self.__filters.append(NodeKindFilter(args, exclude=True))
        return self

    def with_node_kinds(self, *args) -> 'Query':
        """Includes nodes of a specific type. Multiple kinds can be passed in.

        .. versionadded:: 2.4"""
        self.__filters.append(NodeKindFilter(args, exclude=False))
        return self

    def sorted_by_title(self, *, reverse=False) -> 'Query':
        """Sort the entries in this query by the metadata field ``title``."""
        return self.sorted_by_metadata('title', reverse=reverse)

    def sorted_by_date(self, *, reverse=False) -> 'Query':
        """Sort the entries in this query by the metadata field ``date``."""
        return self.sorted_by_metadata('date', reverse=reverse)

    def sorted_by_metadata(self, tag: str, *,
                           reverse=False, case_sensitive=False) -> 'Query':
        """Sort the entries in this query by the specified metadata field."""
        self.__sorters.append(MetadataSorter(tag, reverse, case_sensitive))
        return self

    def reversed(self) -> 'Query':
        """Return the results of this query in reversed order."""
        self.__reversed = True
        return self

    def __execute(self):
        if self.__result is not None:
            return

        self.__result = []
        result = self.__nodes
        for f in self.__filters:
            result: List[Node] = list(filter(lambda x: f.match(x), result))

        if self.__sorters:
            for s in self.__sorters:
                # We ignore this as there's no way to express that `get_key`
                # returns something comparable using built-in types, and for
                # now we want to avoid defining types for the sake of type
                # checking only
                result = sorted(result, key=s.get_key, reverse=s.reverse) # type: ignore

        # Note: Queries can be returned from collections, which are sorted, so
        # not having a sorter specified in this query doesn't imply no specified
        # order
        if self.__reversed:
            result = list(reversed(result))

        def Wrap(n: Node) -> Union[Node, Page]:
            if n.kind in {NodeKind.Document, NodeKind.Index}:
                return Page(n)

            return n

        if self.__limit > 0:
            for i, e in enumerate(result):
                if i >= self.__limit:
                    break
                self.__result.append(Wrap(e))
        else:
            for e in result:
                self.__result.append(Wrap(e))

    def __iter__(self) -> Iterator[Union[Node, Page]]:
        self.__execute()
        assert self.__result is not None
        return iter(self.__result)

    def __len__(self) -> int:
        self.__execute()
        assert self.__result is not None
        return len(self.__result)



from .nodes import DocumentNode
from .site import Site
import pathlib
from typing import Dict, Iterable, List
from enum import Enum, auto
from collections import defaultdict
import logging


def _extract_links(document: DocumentNode):
    """Extract all links from ``<a>`` and ``<img>`` tags in a document.

    This assumes the document has been already processed into valid Html.
    """
    from bs4 import BeautifulSoup
    
    if not document.content:
        # empty document
        return []
    
    soup = BeautifulSoup(document.content, 'lxml')

    for item in soup.find_all(['img', 'a']):
        target = None

        if item.name == 'img':
            target = item.attrs.get('src', None)
        elif item.name == 'a':
            target = item.attrs.get('href', None)

        if target and not target.startswith('#'):
            yield target


class LinkType(Enum):
    Internal = auto()
    External = auto()


def _is_internal_link(link: str) -> bool:
    """Check if a link is internal.

    A link is considered internal if it starts with a single slash (not two,
    as this indicates a link using the same protocol.)
    """
    if len(link) >= 2:
        return link[0] == '/' and link[1] != '/'
    else:
        return link[0] == '/'


def gather_links(documents: Iterable[DocumentNode], link_type: LinkType) \
        -> Dict[str, List[pathlib.PurePosixPath]]:
    """Gather links across documents.

    :return: A dictionary containing a link, and the list of document paths
             in which this link was found.
    """
    result = defaultdict(list)

    for document in documents:
        links = _extract_links(document)
        for link in links:
            if link_type == LinkType.Internal and not _is_internal_link(link):
                continue
            elif link_type == LinkType.External and _is_internal_link(link):
                continue

            result[link].append(document.path)

    return result


def validate_internal_links(links: Dict[str, List[pathlib.PurePosixPath]],
                            site: Site) -> int:
    """Validate internal links.

    For each link, check if it exists inside the provided site. If not, an
    error is printed indicating the link and the documents referencing it."""

    log = logging.getLogger('liara.cmdline.LinkValidator')

    error_count = 0

    for link_str, sources in links.items():
        link = pathlib.PurePosixPath(link_str)

        # Special case handling for index.html:
        # The redirection table contains full paths including the trailing
        # index.html, so if we find a link `/foo/bar/`, we also check
        # `/foo/bar/index.html` in case a redirection is present.
        index_url = link / 'index.html'
        if index_url in site.urls:
            continue

        if link not in site.urls:
            for source in sources:
                log.error(f'"{link}" referenced in "{source}" does not exist')
                error_count += 1

    return error_count


def _check_external_link(url: str):
    """Issue a request to the external URL and check for a valid response.
    """
    import requests
    try:
        ok = False

        if url.startswith('mailto'):
            return (True, url, None,)

        error = 'Unknown error'
        
        try:
            r = requests.get(url, timeout=5)
            if r.status_code == 200:
                ok = True
            else:
                error = f'got {r.status_code}, expected 200'
        except requests.exceptions.ConnectTimeout:
            error = "connection timeout"
        except requests.exceptions.ConnectionError:
            error = "connection error"
        except requests.exceptions.ReadTimeout:
            error = "read timeout"
        except requests.exceptions.TooManyRedirects:
            error = "too many redirects"
        except Exception as e:
            error = str(e)

        if not ok:
            return (False, url, error,)
        else:
            return (True, url, None,)
    except (KeyboardInterrupt, SystemExit):
        return None


def validate_external_links(links: Dict[str, List[pathlib.PurePosixPath]]) \
        -> int:
    """Validate external links.

    This issues a request for each link, and checks if it connects correctly.
    If not, an error is printed indicating the link and the documents
    referencing it."""
    import multiprocessing

    log = logging.getLogger('liara.cmdline.LinkValidator')
    error_count = 0

    try:
        with multiprocessing.Pool() as pool:
            result = pool.imap_unordered(_check_external_link, links.keys())

            for r in result:
                # Only returned when aborting
                if r is None:
                    break

                if not r[0]:
                    for source in links[r[1]]:
                        log.error(
                            f'Link "{r[1]}", referenced in "{source}" '
                            f'failed with: {r[2]}')
                        error_count += 1
    except (KeyboardInterrupt, SystemExit):
        return 0

    return error_count



import cProfile
from . import Liara
from .config import create_default_configuration
from .yaml import dump_yaml
import logging
import os
import click
from .nodes import NodeKind
from typing import Literal, IO, Callable, Optional, Union
from .site import Site
from .nodes import Node, NodeKind, _parse_node_kind
import pathlib
from .cache import NullCache

class Environment:
    """The command line environment.

    This provides access to global variables that are useful for command line
    commands, as well as a global liara instance."""
    __liara : Liara | None

    def __init__(self):
        self.verbose = False
        self.debug = False
        self.config = None
        self.__liara = None
        self.log = logging.getLogger('liara.cmdline')

    @property
    def liara(self) -> Liara:
        if not self.__liara:
            self.__liara = _create_liara(self.config)
        assert self.__liara is not None
        return self.__liara


pass_environment = click.make_pass_decorator(Environment, ensure=True)


def _setup_logging(*, debug: bool, verbose: bool):
    if debug:
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s %(levelname)-7s %(name)s %(message)s')
        # Unfortunately PIL writes a lot of debug output, so we're disabling it
        # manually here to make debug output useful
        # This doesn't remove critical debug output -- PIL writes things like
        # PNG header data here
        logging.getLogger('PIL').setLevel(logging.INFO)
        # Markdown writes a lot of debug output as well, mostly what extensions
        # were loaded and such, which is not important for debugging
        logging.getLogger('MARKDOWN').setLevel(logging.INFO)
    elif verbose:
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)-7s %(message)s')
    else:
        logging.basicConfig(
            level=logging.WARN,
            format='%(asctime)s %(levelname)-7s %(message)s')


@click.group(name='Built-in commands')
@click.option('--debug/--no-debug', default=False, help='Enable debug output.')
@click.option('--verbose', is_flag=True, help='Enable verbose output.')
@click.option('--config', default='config.yaml', type=click.Path(),
              help='Set the path to the configuration file.')
@click.option('--date', default=None, help='Override the current date.')
@click.version_option()
@pass_environment
def cli(env: Environment, debug: bool, verbose: bool, config, date: str):
    """Liara is a static page generator."""
    _setup_logging(debug=debug, verbose=verbose)

    if date:
        from .util import set_local_now
        import dateparser
        now = dateparser.parse(date)
        if now:
            set_local_now(now)
        else:
            env.log.error(f'Could not parse date {date}, ignoring it for this build')

    env.config = config


def main():
    from .signals import commandline_prepared

    Liara.setup_plugins()
    commandline_prepared.send(cli=cli)
    cli()


def _create_liara(config, *, overrides=None):
    if os.path.exists(config):
        return Liara(config, configuration_overrides=overrides)
    else:
        return Liara(configuration_overrides=overrides)


@cli.command()
@click.option('--profile/--no-profile')
@click.option('--profile-file', type=click.Path(writable=True),
              default='build.prof')
@click.option('--cache/--no-cache', default=True,
              help='Enable or disable the configured cache')
@click.option('--parallel/--no-parallel', default=True,
              help='Enable or disable parallel processing.')
@pass_environment
def build(env: Environment, profile: bool,
          profile_file: str | bytes | os.PathLike,
          cache: bool, parallel: bool):
    """Build a site."""
    if profile:
        pr = cProfile.Profile()
        pr.enable()
    env.liara.build(disable_cache=not cache,
                    parallel_build=parallel)
    if profile:
        pr.disable()
        pr.dump_stats(profile_file)


@cli.command()
@click.option('--type', '-t', 'link_type',
              type=click.Choice(['internal', 'external']),
              default='internal')
@pass_environment
def validate_links(env: Environment, link_type: Literal['internal', 'external']):
    """Validate links.

    Checks all internal/external links for validity. For internal links,
    a check is performed if the link target exists. For external links,
    a web request is made to the link target. If the request returns an
    error code, the link is considered invalid."""
    from .cache import MemoryCache
    from .actions import (
        validate_internal_links,
        validate_external_links,
        gather_links,
        LinkType
    )
    liara = env.liara
    site = liara.discover_content()
    cache = MemoryCache()

    env.log.debug('Processing site ...')
    args = {
        '$data': site.merged_data
    }
    for document in site.documents:   
        document.process(cache, **args)

    env.log.debug('done')

    if link_type == 'internal':
        link_type = LinkType.Internal
        env.log.debug('Checking internal links')
    elif link_type == 'external':
        link_type = LinkType.External
        env.log.debug('Checking external links')

    links = gather_links(site.documents, link_type)
    env.log.debug(f'Found {len(links)} {link_type.name.lower()} links')

    error_count = 0

    if link_type == LinkType.Internal:
        error_count = validate_internal_links(links, site)
    elif link_type == LinkType.External:
        error_count = validate_external_links(links)

    if error_count > 0:
        env.log.error(f'Found {error_count} broken links')

        return 1


@cli.command()
@pass_environment
def list_tags(env: Environment):
    """List all tags.

    This uses a metadata field named 'tags' and returns the union of all tags,
    as well as the count how often each tag is used."""
    from collections import Counter
    liara = env.liara
    site = liara.discover_content()
    tags = []
    for document in site.documents:
        tags += document.metadata.get('tags', [])

    counted_tags = sorted(Counter(tags).items(), key=lambda x: x[1],
                          reverse=True)
    for k, v in counted_tags:
        print(k, v)


@cli.command()
@click.argument('tags', nargs=-1)
@pass_environment
def find_by_tag(env: Environment, tags: list[str]):
    """Find pages by tag.

    This searches the metadata for a 'tags' field, which is assumed to be
    a list of tags."""
    liara = env.liara
    site = liara.discover_content()
    tags = set(tags)
    for document in site.documents:
        for tag in document.metadata.get('tags', []):
            if tag in tags:
                print(document.src, tag)
                break


@cli.command()
@click.option('--type', '-t', 'object_type')
@pass_environment
def create(env: Environment, object_type: str):
    """Create a document."""
    liara = env.liara
    liara.discover_content()
    liara.create_document(object_type)


@cli.command()
@click.option('--output', '-o', type=click.File(mode='w'))
def create_config(output: IO):
    """Create a default configuration."""
    dump_yaml(create_default_configuration(), output)


@cli.command()
@click.option('--template-backend', '-t',
              type=click.Choice(['jinja2', 'mako']),
              default='jinja2')
def quickstart(template_backend: Literal['jinja2', 'mako']):
    """Create a quickstart project."""
    from .quickstart import generate
    generate(template_backend)


class _Node:
    """Helper class for tree printing, as the liara site node tree doesn't
    contain intermediate nodes."""
    def __init__(self, name: str, data=None):
        self.__name = name
        self.__children = []
        self.__data = data

    def add_child(self, node: "_Node"):
        self.__children.append(node)

    @property
    def children(self) -> list["_Node"]:
        return self.__children

    @property
    def name(self):
        return self.__name
    
    @property
    def data(self):
        return self.__data


def _print_tree(node: _Node, get_label: Callable[[_Node], str], prefix='', last=True):
    """Pretty-print a fully populated tree (meaning: all intermediate nodes
    are present.)"""
    empty = "    "
    last_branch = "└── "
    continue_traversal = "│   "
    branch = "├── "
    
    if node.name == 'Site':
        # Special case the root node: Don't add a prefix here
        print('Site')
        if node.children:
            for i, c in enumerate(sorted(node.children, key=lambda n: n.name)):
                _print_tree(c, get_label, '', i == (len(node.children) - 1))
    else:
        # Normal node - print the prefix so far, append the right branch
        # symbol, then the label
        print(prefix + (last_branch if last else branch) + get_label(node))
        if node.children:
            for i, c in enumerate(sorted(node.children, key=lambda n: n.name)):
                _print_tree(c, get_label, 
                            # Logic here is: If we're last we're not adding
                            # a vertical bar, just spaces
                            prefix + (empty if last else continue_traversal),
                            i == (len(node.children) - 1))


@cli.command()
@click.option('--format', '-f', type=click.Choice(['tree', 'list', 'json']),
              default='tree')
@click.option('--type', '-t', 'content_type', multiple=True,
              type=click.Choice(
                  list(map(str.lower, NodeKind.__members__.keys()))))
@pass_environment
def list_content(env: Environment,
                 format: Literal['tree', 'list', 'json'],
                 content_type: list[str]):
    """List all content.

    If ``format`` is set to ``tree``, this will print the content as a
    tree. If ``format`` is set to ``list``, this will produce a
    flat list instead. JSON will print a JSON structure with information for
    each node.
    """
    liara = env.liara
    content = liara.discover_content()

    # We sort by path name, which makes it trivial to sort it later into a tree
    # as children always come after their parent
    nodes = sorted(content.nodes, key=lambda x: x.path)
    if not nodes:
        return

    def filter_nodes(node):
        if node.kind.name.lower() not in content_type:
            return False
        return True

    if content_type:
        nodes = list(filter(filter_nodes, nodes))

    def get_node_label(node):
        label = f"{node.path.parts[-1]}"
        if node.kind == NodeKind.Resource:
            label += f"(Resource, generated from '{node.src}')"
        else:
            label +=  f"({node.kind.name})"
        
        return label

    if format == 'tree':
        root = _Node('Site')
        node_map = {'/': root}

        def add_or_create(path, data=None):
            if str(path) in node_map:
                return

            # Create parent nodes recursively, as needed
            add_or_create(path.parent)
            
            n = _Node(path.name, data)
            node_map[str(path.parent)].add_child(n)
            node_map[str(path)] = n
        
        for node in nodes:
            path = node.path

            add_or_create(path, node)

        import sys
        # This seems to be required to get UTF-8 output redirection to work
        # in powershell. Unclear why
        sys.stdout.reconfigure(encoding='utf-8')

        def get_label(node):
            if node.data:
                return get_node_label(node.data)
            return node.name

        _print_tree(root, get_label)
    elif format == 'list':
        for node in nodes:
            print(str(node.path), get_node_label(node))
    elif format == 'json':
        import json
        result = {
            'version': 1,
            'nodes': []
        }
        for node in nodes:
            data = {
                'path': str(node.path),
                'kind': node.kind.name
            }
            if node.src:
                data['source'] = str(node.src)

            result['nodes'].append(data)
        print(json.dumps(result, indent=4))


@cli.command()
@click.option('--browser/--no-browser',
              default=True,
              help='Open a browser window automatically')
@click.option('--port', '-p', default=8080,
              help='The port to use for the local server')
@click.option('--cache/--no-cache', default=True,
              help='Enable or disable the configured cache')
@pass_environment
def serve(env: Environment, browser: bool, port: int, cache: bool):
    """Run a local development server."""
    import watchfiles
    from .server import HttpServer

    liara = _create_liara(env.config)
    # The HTTP server runs a thread internally
    server = HttpServer(open_browser=browser, port=port)

    liara._set_base_url_override(server.get_url())

    server.start(liara,
                 liara._get_cache() if cache else NullCache())

    def get_ignore_directories(config):
        output_directory = config["output_directory"]
        ignore_dirs = [output_directory]

        match config["build.cache.type"]:
            case "fs":
                ignore_dirs.append(config["build.cache.fs.directory"])
            case "db":
                ignore_dirs.append(config["build.cache.db.directory"])
        
        return ignore_dirs

    try:
        # We need this nested loop in case the cache or output directory
        # changes. The server can actually cope with this, but we need to
        # ignore those folders
        while True:
            config = liara._get_configuration()
            ignore_dirs = get_ignore_directories(config)

            watch_filter = watchfiles.DefaultFilter(ignore_dirs=ignore_dirs)
            
            for change in watchfiles.watch('.',
                                        watch_filter=watch_filter,
                                        recursive=True):
                # TODO We can optimize this by skipping  the reconfigure
                # if it's a _known_ content node that changes

                # This is needed to bring the cache into a sensible state again,
                # for example, the SQLite cache would otherwise keep a
                # transaction going
                liara._get_cache().persist()

                liara = _create_liara(env.config)
                liara._set_base_url_override(server.get_url())

                server.reconfigure(liara,
                                   liara._get_cache() if cache else NullCache())

                # If the directories didn't change, there's no need to setup
                # the watch again, so we skip this and return to the loop early
                new_ignore_dirs = get_ignore_directories(liara._get_configuration())
                if new_ignore_dirs != ignore_dirs:
                    break
    except KeyboardInterrupt:
        pass

    server.stop()
    liara._get_cache().persist()


@cli.command()
@click.argument('action', type=click.Choice(['clear', 'inspect']))
@pass_environment
def cache(env: Environment, action: Literal['clear', 'inspect']):
    """Modify or inspect the cache."""
    import humanfriendly

    if action == 'clear':
        env.liara._get_cache().clear()
    elif action == 'inspect':
        info = env.liara._get_cache().inspect()
        size = humanfriendly.format_size(info.size, binary=True)
        print(f'Cache type:   {info.name}')
        print(f'Size:         {size}')
        print(f'Object count: {info.entry_count}')

@cli.group()
def inspect():
    """Inspect state, for example, site data or template matching."""
    pass

@inspect.command('data')
@pass_environment
def inspect_data(env: Environment):
    """Inspect the content of `site.data`."""
    import yaml
    liara = env.liara
    liara.discover_content()

    # Using dict() here improves formatting, as pprint starts with
    # mappingproxy(...) otherwise which is an implementation detail
    print(yaml.dump(dict(liara.site.merged_data),
                    default_flow_style=False,
                    allow_unicode=True))


@inspect.command('list-data-files')
@pass_environment
def inspect_data_files(env: Environment):
    """List all data files."""
    liara = env.liara
    liara.discover_content()

    for df in liara.site.data:
        print(df.src)

class _SiteShim(Site):
    """Helper class to inject fake nodes for `template-match` so
    we can match selectors with `?kind=` specified."""
    def __init__(self, site: Site):
        self.__site = site
        self.__fakes = {}

    def register_fake_node(self, path: str, kind: NodeKind):
        self.__fakes[pathlib.PurePosixPath(path)] = kind

    def get_node(self, path: Union[str, pathlib.PurePosixPath]) \
            -> Optional[Node]:
        if path in self.__fakes:
            assert isinstance(path, pathlib.PurePosixPath)
            n = Node()
            n.path = path
            match self.__fakes[path]:
                case NodeKind.Document:
                    n.kind = NodeKind.Document
                case NodeKind.Index:
                    n.kind = NodeKind.Index

            return n
                    
        return self.__site.get_node(path)

@inspect.command('template-match')
@click.argument('path')
@click.option('--kind', type=click.Choice(['document', 'index']),
              help='Override or set the node kind at the specified path.')
@pass_environment
def inspect_template_match(env: Environment, path: str, kind: Optional[Literal['document', 'index']]):
    """Match a template pattern."""
    liara = env.liara
    liara.discover_content()

    if not path.startswith('/'):
        path = '/' + path

    site_shim = _SiteShim(liara.site)
    if kind is not None:
        site_shim.register_fake_node(path, _parse_node_kind(kind))
    else:
        if (node := liara.site.get_node(path)) is None:
            site_shim.register_fake_node(path, NodeKind.Document)
        else:
            print(f'Node "{path}" exists with kind: {node.kind.name}')

    template_repository = liara._get_template_repository()
    match, pattern = template_repository._match_template(pathlib.PurePosixPath(path), site_shim)
    print(f'Using template "{match}" for path "{path}" due to pattern "{pattern}"')


if __name__ == '__main__':
    main()



from .yaml import dump_yaml
from .config import create_default_metadata

__TEMPLATES_JINJA2 = {
    'page.jinja2': r"""
<!DOCTYPE html>
<html>
<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
    <meta name=viewport content="width=device-width, initial-scale=1">
      <title>
      {% block title %}
          {% if page.metadata.title and page.url != '/' %}
            {{ page.metadata.title }}
        {% else %}
            My blog
        {% endif %}
        {% endblock %}
    </title>
    <link type="text/css" rel="stylesheet" href="/style.css" />
</head>
<body>
    <div class="container">
        <header id="header">
            <nav>
                <a href="/">Home</a>
                <a href="/archive">Archive</a>
            </nav>
        </header>
        <main id="content">
        {% block content %}
            {{ page.content }}
        {% endblock %}
        </main>
        <footer id="footer">
            <p>Generated using Liara</p>
        </footer>
    </div>
</body>
</html>
""",
    'blog.jinja2': r"""{% extends "page.jinja2" %}
{% block content %}
<article>
    <header>
        <h1>{{page.meta.title}}</h1>
        <div class="metadata">
            <div class="post-time">
              <time datetime="{{ page.metadata.date }}">{{
                  page.metadata.date.strftime("%Y-%m-%d, %H:%M") }}</time>
            </div>
            <div class="tags">
              <ul class="tags">
                {% for tag in page.metadata.tags|sort %}
                  <li>
                    <a href="/archive/by-tag/{{ tag }}">
                        {{ tag|capitalize }}
                    </a>
                   </li>
                {% endfor %}
              </ul>
            </div>
        </div>
    </header>
    {{ page.content }}
    <aside>
        <div class="pagination">
            {% set previous = site.get_previous_in_collection('blog', page) %}
            {% set next = site.get_next_in_collection('blog', page) %}
            {% if previous %}
                <div class="previous">
                    <a href="{{ previous.url }}">Previous post</a>
                </div>
            {% endif %}
            {% if next %}
                <div class="next">
                    <a href="{{ next.url }}">Next post</a>
                </div>
            {% endif %}
        </div>
    </aside>
</article>
{% endblock %}
""",
    'archive.jinja2': r"""
{% extends "page.jinja2" %}
{% block content %}
<h1>Blog archive {% if page.metadata.key %}for {{ page.metadata.key }}{% endif %}</h1>
{{ page.content }}
<ul>
    {% if not page.metadata.top_level_index %}
        {% for ref in page.references %}
        <li>
            <a href="{{ ref.url }}">{{ ref.meta.title }}</a>
        </li>
        {% endfor %}
    {% else %}
        {% for child_page in
        page.children.sorted_by_metadata('key') %}
        <li>
            <a href="{{ child_page.url }}">{{ child_page.metadata.key }}</a>
            ({{ child_page.references|length }} posts)
            <ul>
                {% for ref in child_page.references.sorted_by_date(reverse=True) %}
                <li>
                    <a href="{{ ref.url }}">{{ ref.meta.title}}</a>
                </li>
                {% endfor %}
            </ul>
        </li>
        {% endfor %}
    {% endif %}
</ul>
{% endblock %}
"""
}

__TEMPLATES_MAKO = {
    'page.mako': r"""
<!DOCTYPE html>
<html>
<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
    <meta name=viewport content="width=device-width, initial-scale=1">
      <title>
        <%block name="title">
          % if page.metadata.get('title') and page.url != '/':
            ${page.metadata['title']}
          % else:
            My blog
          % endif
        </%block>
    </title>
    <link type="text/css" rel="stylesheet" href="/style.css" />
</head>
<body>
    <div class="container">
        <header id="header">
            <nav>
                <a href="/">Home</a>
                <a href="/archive">Archive</a>
            </nav>
        </header>
        <main id="content">
        <%block name="content">
            ${page.content}
        </%block>
        </main>
        <footer id="footer">
            <p>Generated using Liara</p>
        </footer>
    </div>
</body>
</html>""",
    'blog.mako': r"""<%inherit file="page.mako"/>
<%block name="content">
<article>
    <header>
        <h1>${page.metadata["title"]}</h1>
        <div class="metadata">
            <div class="post-time">
              <time datetime="${page.metadata['date']}">
                  ${page.metadata['date'].strftime("%Y-%m-%d, %H:%M")}</time>
            </div>
            <div class="tags">
              <ul class="tags">
                % for tag in sorted(page.metadata['tags']):
                  <li>
                    <a href="/archive/by-tag/${tag}">
                        ${tag.capitalize()}
                    </a>
                   </li>
                % endfor
              </ul>
            </div>
        </div>
    </header>
    ${page.content}
    <aside>
        <div class="pagination">
            <%
                previous = site.get_previous_in_collection('blog', page)
                next = site.get_next_in_collection('blog', page)
            %>
            % if previous:
                <div class="previous">
                    <a href="${previous.url}">Previous post</a>
                </div>
            % endif
            % if next:
                <div class="next">
                    <a href="${next.url}">Next post</a>
                </div>
            % endif
        </div>
    </aside>
</article>
</%block>""",
    'archive.mako': r"""

<%inherit file="page.mako"/>
<%block name="content">
<h1>Blog archive
% if page.metadata.get('key'):
    for ${page.metadata['key']}
% endif
</h1>
${page.content}
<ul>
    % if not page.metadata.get('top_level_index', False):
        % for ref in page.references:
        <li>
            <a href="${ref.url}">${ref.metadata['title']}</a>
        </li>
        % endfor
    % else:
        % for child_page in page.children.sorted_by_metadata('key'):
        <li>
            <a href="${child_page.url}">${child_page.metadata['key']}</a>
            (${len(child_page.references)} posts)
            <ul>
                % for ref in child_page.references.sorted_by_date(reverse=True):
                <li>
                    <a href="${ref.url}">${ref.metadata['title']}</a>
                </li>
                % endfor
            </ul>
        </li>
        % endfor
    % endif
</ul>
</%block>"""
}

__SCSS = """
.container {
    max-width: 800px;

    margin: auto;

    header {
        background-color: #EEE;
        padding: 0.5em;
    }

    main {
        padding: 0.5em;

        header {
            padding: 0;
            background-color: unset;
        }
    }

    footer {
        font-size: 0.8em;
        padding: 0.5em;
        border-top: 1px solid #CCC;
    }
}
"""

__THEME_CONFIG_JINJA2 = {
    'backend': 'jinja2',
    'paths': {
        '/blog/*?kind=document': 'blog.jinja2',
        '/archive/*': 'archive.jinja2',
        '/*': 'page.jinja2',
    },
    'resource_directory': 'resources'
}

__THEME_CONFIG_MAKO = {
    'backend': 'mako',
    'paths': {
        '/blog/*?kind=document': 'blog.mako',
        '/archive/*': 'archive.mako',
        '/*': 'page.mako',
    },
    'resource_directory': 'resources'
}

__BLOG_POST_GENERATOR = r"""
import liara
import pathlib
import datetime


def generate(site: liara.site.Site, configuration) -> pathlib.Path:
    title = input("Blog title: ")
    slug = liara.util.create_slug(title)
    tags = input("Tags (comma separated): ")
    # need list to avoid YAML containing special tags
    tags = list(set(map(str.strip, tags.split(','))))
    now = liara.util.local_now()

    metadata = liara.yaml.dump_yaml({
        'title': title,
        'tags': tags,
        'date': now
    })

    content = '---\n'
    content += metadata
    content += '---\n'
    content += '\n'
    content += 'Blog post content goes here!'

    content_directory = configuration['content_directory']
    output_file = pathlib.Path(f"{content_directory}/blog/{now.year}/{slug}.md")
    output_file.parent.mkdir(parents=True, exist_ok=True)
    output_file.write_text(content)
    return output_file
"""


def generate_templates(backend):
    if backend == 'jinja2':
        templates = __TEMPLATES_JINJA2
    elif backend == 'mako':
        templates = __TEMPLATES_MAKO

    for k, v in templates.items():
        open(f'templates/{k}', 'w').write(v)


def generate_css():
    import os
    os.makedirs('templates/resources', exist_ok=True)
    open('templates/resources/style.scss', 'w').write(__SCSS)


def generate_theme(backend):
    import os
    os.makedirs('templates', exist_ok=True)
    generate_templates(backend)
    generate_css()

    if backend == 'jinja2':
        theme_config = __THEME_CONFIG_JINJA2
    elif backend == 'mako':
        theme_config = __THEME_CONFIG_MAKO
    dump_yaml(theme_config, open('templates/default.yaml', 'w'))


def generate_content():
    import datetime
    import os
    page_template = """---
title: %TITLE%
tags: [%TAGS%]
date: %DATE%
---

Sample post using _Markdown_
"""

    dates = [
        datetime.datetime(2019, 1, 3),
        datetime.datetime(2018, 1, 5),
        datetime.datetime(2017, 3, 1)
    ]

    titles = [
        'Newest post',
        'Old but not oldest post',
        'The beginning'
    ]

    tags = [
        ['new', 'featured'],
        ['old'],
        ['old', 'historic']
    ]

    slugs = [
        'newest-post',
        'old-but-not-oldest-post',
        'the-beginning'
    ]

    os.makedirs('content/blog', exist_ok=True)

    for date, title, tag, slug in zip(dates, titles, tags, slugs):
        t = page_template.replace("%TITLE%", title)
        t = t.replace("%TAGS%", ', '.join(tag))
        t = t.replace("%DATE%", str(date))
        os.makedirs(f'content/blog/{date.year}', exist_ok=True)
        open(f'content/blog/{date.year}/{slug}.md', 'w').write(t)

    open('content/archive.md', 'w').write("""
---
title: The archive
---

The blog archive. Find blog posts [by year](/archive/by-year) or
[by tag](/archive/by-tag).""")

    open('content/_index.md', 'w').write("""
---
title: Hello world
---

Welcome to the sample blog.""")


__DEFAULT_CONFIG = {
    'collections': 'collections.yaml',
    'indices': 'indices.yaml',
}

__DEFAULT_COLLECTIONS = {
    'blog': {
        'filter': '/blog/**',
        'order_by': 'date'
    }
}

__DEFAULT_INDICES = [
    {
        'collection': 'blog',
        'group_by': ['date.year'],
        'path': '/archive/by-year/%1',
        'create_top_level_index': True
    },
    {
        'collection': 'blog',
        'group_by': ['*tags'],
        'path': '/archive/by-tag/%1',
        'create_top_level_index': True
    }
]


def generate_configs():
    dump_yaml(__DEFAULT_CONFIG, open('config.yaml', 'w'))
    dump_yaml(create_default_metadata(), open('metadata.yaml', 'w'))
    dump_yaml(__DEFAULT_COLLECTIONS, open('collections.yaml', 'w'))
    dump_yaml(__DEFAULT_INDICES, open('indices.yaml', 'w'))


def generate_generator():
    import os
    os.makedirs('generators', exist_ok=True)
    output_file = os.path.join('generators', 'blog-post.py')
    open(output_file, 'w').write(__BLOG_POST_GENERATOR)


def generate(template_backend):
    generate_theme(template_backend)
    generate_content()
    generate_configs()
    generate_generator()



from blinker import signal


content_filtered = signal('content-filtered')
"""Raised when content has been removed due to a filter.

  This signal is raised during the content discovery stage.

  :param liara.nodes.Node node: the node that was removed
  :param liara.site.ContentFilter filter: the filter that matched
"""

content_added = signal('content-added')
"""Raised when a content node was successfully added.

  This signal is raised during the content discovery stage. It is not raised
  for nodes that have been filtered.

  :param liara.nodes.Node node: the node that was created
"""

commandline_prepared = signal('commandline-prepared')
"""Raised when the command line parsing environment was prepared.

  :param click.group cli: The command line group to add commands to.
"""

content_discovered = signal('content-discovered')
"""Raised after all content has been discovered.

  :param liara.site.Site site: the site instance
"""

documents_processed = signal('documents-processed')
"""Raised after all documents have been processed.

  :param liara.site.Site site: the site instance

  Processing includes the conversion from Markdown to HTML.
"""

document_loaded = signal('document-loaded')
"""Raised after a document has been loaded.

  This signal is raised during the content discovery stage.

  :param liara.nodes.DocumentNode document: the document node
  :param str content: the raw document contents

  When this signal is raised, the content has been loaded, but no
  templates etc. have been applied to the document yet.
"""

register_markdown_shortcodes = signal('register-markdown-shortcodes')
"""Raised while initializing the Markdown processor.

  :param liara.md.ShortcodePreprocessor preprocessor: The shortcode
                                                      preprocessor

  When this signal is raised, you can register shortcode function calls by
  calling :py:meth:`liara.md.ShortcodePreprocessor.register`.
"""
